"""Automatically generated module split."""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from pyrung.core.tag import Tag

from .base import Instruction, OneShotMixin, SubroutineReturnSignal
from .conversions import (
    _store_copy_value_to_tag_type,
)
from .resolvers import (
    _fn_name,
    resolve_tag_ctx,
    resolve_tag_or_value_ctx,
)
from .utils import guard_oneshot_execution

if TYPE_CHECKING:
    from pyrung.core.condition import Condition
    from pyrung.core.context import ScanContext
    from pyrung.core.memory_block import IndirectExprRef, IndirectRef


class FunctionCallInstruction(OneShotMixin, Instruction):
    """Stateless function call: copy-in / execute / copy-out."""

    def __init__(
        self,
        fn: Callable[..., dict[str, Any]],
        *,
        ins: dict[str, Tag | IndirectRef | IndirectExprRef | Any] | None = None,
        outs: dict[str, Tag | IndirectRef | IndirectExprRef] | None = None,
        oneshot: bool = False,
    ):
        OneShotMixin.__init__(self, oneshot)
        self._fn = fn
        self._ins = ins or {}
        self._outs = outs or {}

    @guard_oneshot_execution
    def execute(self, ctx: ScanContext, enabled: bool) -> None:
        kwargs = {name: resolve_tag_or_value_ctx(src, ctx) for name, src in self._ins.items()}
        result = self._fn(**kwargs)
        if not self._outs:
            return
        if result is None:
            raise TypeError(
                f"run_function: {_fn_name(self._fn)!r} returned None but outs were declared"
            )
        for key, target in self._outs.items():
            if key not in result:
                raise KeyError(
                    f"run_function: {_fn_name(self._fn)!r} missing key {key!r}; got {sorted(result)}"
                )
            resolved = resolve_tag_ctx(target, ctx)
            ctx.set_tag(resolved.name, _store_copy_value_to_tag_type(result[key], resolved))


class EnabledFunctionCallInstruction(Instruction):
    """Always-execute function call with enabled flag."""

    ALWAYS_EXECUTES = True
    INERT_WHEN_DISABLED = False

    def __init__(
        self,
        fn: Callable[..., dict[str, Any]],
        *,
        ins: dict[str, Tag | IndirectRef | IndirectExprRef | Any] | None = None,
        outs: dict[str, Tag | IndirectRef | IndirectExprRef] | None = None,
        enable_condition: Condition | None,
    ):
        self._fn = fn
        self._ins = ins or {}
        self._outs = outs or {}
        self._enable_condition = enable_condition

    def execute(self, ctx: ScanContext, enabled: bool) -> None:
        kwargs = {name: resolve_tag_or_value_ctx(src, ctx) for name, src in self._ins.items()}
        result = self._fn(enabled, **kwargs)
        if not self._outs:
            return
        if result is None:
            raise TypeError(
                f"run_enabled_function: {_fn_name(self._fn)!r} returned None but outs were declared"
            )
        for key, target in self._outs.items():
            if key not in result:
                raise KeyError(
                    "run_enabled_function: "
                    f"{_fn_name(self._fn)!r} missing key {key!r}; got {sorted(result)}"
                )
            resolved = resolve_tag_ctx(target, ctx)
            ctx.set_tag(resolved.name, _store_copy_value_to_tag_type(result[key], resolved))


class ForLoopInstruction(OneShotMixin, Instruction):
    """For-loop instruction.

    Executes a captured instruction list N times within one scan.
    """

    def __init__(
        self,
        count: Tag | IndirectRef | IndirectExprRef | Any,
        idx_tag: Tag,
        instructions: list[Instruction],
        oneshot: bool = False,
    ):
        OneShotMixin.__init__(self, oneshot)
        self.count = count
        self.idx_tag = idx_tag
        self.instructions = instructions

    def execute(self, ctx: ScanContext, enabled: bool) -> None:
        if not enabled:
            for instruction in self.instructions:
                instruction.execute(ctx, False)
            self.reset_oneshot()
            return

        if not self.should_execute(enabled):
            return

        count_value = resolve_tag_or_value_ctx(self.count, ctx)
        iterations = max(0, int(count_value))

        for i in range(iterations):
            # Keep loop index in tag space so indirect refs resolve via ctx.get_tag().
            ctx.set_tag(self.idx_tag.name, i)
            for instruction in self.instructions:
                instruction.execute(ctx, True)

    def reset_oneshot(self) -> None:
        """Reset own oneshot state and propagate reset to captured children."""
        OneShotMixin.reset_oneshot(self)
        for instruction in self.instructions:
            reset_fn = getattr(instruction, "reset_oneshot", None)
            if callable(reset_fn):
                reset_fn()


class CallInstruction(Instruction):
    """Call subroutine instruction.

    Executes a named subroutine when the rung is true.
    The subroutine must be defined in the same Program.
    """

    def __init__(self, subroutine_name: str, program: Any):
        self.subroutine_name = subroutine_name
        self._program = program  # Reference to Program for subroutine lookup

    def execute(self, ctx: ScanContext, enabled: bool) -> None:
        if not enabled:
            return
        self._program.call_subroutine_ctx(self.subroutine_name, ctx)


class ReturnInstruction(Instruction):
    """Return from the current subroutine immediately."""

    def execute(self, ctx: ScanContext, enabled: bool) -> None:  # noqa: ARG002
        if not enabled:
            return
        raise SubroutineReturnSignal



"""Shared structures and errors for Click ladder CSV export."""

from __future__ import annotations

import csv
import logging
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from pyrung.core.condition import Condition

# ---- Shared issue payload ----
Issue = dict[str, str | int | None]


# ---- Public/raised errors ----
class LadderExportError(RuntimeError):
    """Raised when strict ladder export prevalidation or lowering fails."""

    def __init__(self, issues: list[Issue] | tuple[Issue, ...]):
        def _safe_int(value: Any) -> int | None:
            if value is None:
                return None
            try:
                return int(value)
            except (TypeError, ValueError):
                return None

        normalized: list[Issue] = []
        for issue in issues:
            normalized.append(
                {
                    "path": str(issue.get("path", "")),
                    "message": str(issue.get("message", "")),
                    "source_file": (
                        None if issue.get("source_file") is None else str(issue.get("source_file"))
                    ),
                    "source_line": _safe_int(issue.get("source_line")),
                }
            )
        self.issues: tuple[Issue, ...] = tuple(normalized)

        if self.issues:
            preview = "; ".join(f"{issue['path']}: {issue['message']}" for issue in self.issues[:3])
            if len(self.issues) > 3:
                preview += f" (+{len(self.issues) - 3} more)"
        else:
            preview = "Ladder export failed."
        super().__init__(preview)


# ---- Export summary ----
@dataclass(frozen=True)
class ExportSummary:
    """Summary of transformations applied during ladder export."""

    renames: tuple[tuple[str, str], ...]  # (dsl_name, csv_name) pairs
    added_next: int  # number of for-loops closed with next()
    added_return: int  # number of subroutines given a return() tail
    added_end: bool  # whether end() was appended to main

    def summary(self) -> str:
        """Render a human-readable summary string."""
        lines: list[str] = []
        if self.renames:
            parts = [f"{dsl} \u2192 {csv}" for dsl, csv in self.renames]
            lines.append(f"Renamed: {', '.join(parts)}")
        added: list[str] = []
        if self.added_next:
            s = "s" if self.added_next != 1 else ""
            added.append(f"next() closing {self.added_next} for-loop{s}")
        if self.added_return:
            s = "s" if self.added_return != 1 else ""
            added.append(f"return() on {self.added_return} subroutine{s}")
        if self.added_end:
            added.append("end() on main")
        if added:
            lines.append(f"Added:   {', '.join(added)}")
        return "\n".join(lines)

    def __str__(self) -> str:
        return self.summary()


# ---- Export payload ----
@dataclass(frozen=True)
class LadderBundle:
    """Row-matrix CSV payload for Click ladder export."""

    main_rows: tuple[tuple[str, ...], ...]
    subroutine_rows: tuple[tuple[str, tuple[tuple[str, ...], ...]], ...]
    export_summary: ExportSummary = field(
        default_factory=lambda: ExportSummary(
            renames=(), added_next=0, added_return=0, added_end=False
        )
    )

    def write(self, directory: str | Path) -> None:
        output_dir = Path(directory)
        output_dir.mkdir(parents=True, exist_ok=True)

        with (output_dir / "main.csv").open("w", encoding="utf-8", newline="") as handle:
            writer = csv.writer(handle)
            writer.writerows(self.main_rows)

        if self.subroutine_rows:
            subroutine_dir = output_dir / "subroutines"
            subroutine_dir.mkdir(parents=True, exist_ok=True)

            slug_counts: dict[str, int] = {}
            for subroutine_name, rows in self.subroutine_rows:
                base_slug = _slugify(subroutine_name)
                count = slug_counts.get(base_slug, 0)
                slug_counts[base_slug] = count + 1
                # Keep filenames unique when multiple subroutines slugify to the same value.
                slug = base_slug if count == 0 else f"{base_slug}_{count + 1}"

                with (subroutine_dir / f"{slug}.csv").open(
                    "w", encoding="utf-8", newline=""
                ) as handle:
                    writer = csv.writer(handle)
                    writer.writerows(rows)

        summary_text = self.export_summary.summary()
        if summary_text:
            logging.getLogger("pyrung.click.ladder").info(summary_text)


class _RenderError(RuntimeError):
    def __init__(self, issue: Issue):
        self.issue = issue
        super().__init__(f"{issue.get('path')}: {issue.get('message')}")


@dataclass
class _ConditionRow:
    cells: list[str]
    cursor: int
    accepts_terms: bool = True

    def clone(self) -> _ConditionRow:
        return _ConditionRow(
            cells=self.cells.copy(),
            cursor=self.cursor,
            accepts_terms=self.accepts_terms,
        )


@dataclass(frozen=True)
class _OutputSlot:
    output_token: str
    local_conditions: tuple[Condition, ...] = ()


# ---- Internal helpers ----
def _slugify(name: str) -> str:
    # Insert underscores at CamelCase boundaries (e.g. AlarmHandler → Alarm_Handler)
    s = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", name)
    s = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", s)
    slug = re.sub(r"[^a-zA-Z0-9]+", "_", s).strip("_").lower()
    return slug if slug else "subroutine"


__all__ = [
    "ExportSummary",
    "Issue",
    "LadderBundle",
    "LadderExportError",
    "_ConditionRow",
    "_OutputSlot",
    "_RenderError",
]



"""P1AM-200 hardware configuration model.

Provides the :class:`P1AM` class which maps physical I/O module slots to
pyrung :class:`~pyrung.core.memory_block.InputBlock` and
:class:`~pyrung.core.memory_block.OutputBlock` instances.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Final, Literal, TypeAlias, overload

from pyrung.circuitpy.catalog import (
    MODULE_CATALOG,
    ModuleSpec,
)
from pyrung.core.memory_block import InputBlock, OutputBlock

MAX_SLOTS: Final[int] = 15
"""Maximum number of I/O module slots on the P1AM-200 base unit."""

InputModuleName: TypeAlias = Literal[
    "P1-08ND-TTL",
    "P1-08ND3",
    "P1-08NA",
    "P1-08SIM",
    "P1-08NE3",
    "P1-16ND3",
    "P1-16NE3",
    "P1-04AD",
    "P1-04AD-1",
    "P1-04AD-2",
    "P1-04RTD",
    "P1-04THM",
    "P1-04NTC",
    "P1-04ADL-1",
    "P1-04ADL-2",
    "P1-08ADL-1",
    "P1-08ADL-2",
]

OutputModuleName: TypeAlias = Literal[
    "P1-04TRS",
    "P1-08TA",
    "P1-08TRS",
    "P1-16TR",
    "P1-08TD-TTL",
    "P1-08TD1",
    "P1-08TD2",
    "P1-15TD1",
    "P1-15TD2",
    "P1-04DAL-1",
    "P1-04DAL-2",
    "P1-08DAL-1",
    "P1-08DAL-2",
]

ComboModuleName: TypeAlias = Literal[
    "P1-16CDR",
    "P1-15CDD1",
    "P1-15CDD2",
    "P1-4ADL2DAL-1",
    "P1-4ADL2DAL-2",
]

SlotValue: TypeAlias = InputBlock | OutputBlock | tuple[InputBlock, OutputBlock]


def _make_formatter(prefix: str) -> Callable[[str, int], str]:
    """Create an address formatter that produces ``"Prefix.N"`` tag names."""

    def formatter(block_name: str, addr: int) -> str:
        return f"{prefix}.{addr}"

    return formatter


class P1AM:
    """P1AM-200 hardware configuration.

    Represents a P1AM-200 base unit with up to 15 I/O module slots.
    Each slot is configured with a module part number via :meth:`slot`,
    which constructs and returns the appropriate
    :class:`~pyrung.core.memory_block.InputBlock` /
    :class:`~pyrung.core.memory_block.OutputBlock`.

    This class holds no runtime state — it produces Block instances
    whose tags reference values in ``SystemState.tags`` via the core engine.

    Example::

        hw = P1AM()
        inputs  = hw.slot(1, "P1-08SIM")        # InputBlock
        outputs = hw.slot(2, "P1-08TRS")         # OutputBlock
        inp, out = hw.slot(3, "P1-16CDR")        # combo → tuple

        Button = inputs[1]    # LiveInputTag("Slot1.1", BOOL)
        Light  = outputs[1]   # LiveOutputTag("Slot2.1", BOOL)
    """

    def __init__(self) -> None:
        self._slots: dict[int, tuple[ModuleSpec, SlotValue]] = {}

    @overload
    def slot(
        self,
        number: int,
        module: InputModuleName,
        *,
        name: str | None = None,
    ) -> InputBlock: ...

    @overload
    def slot(
        self,
        number: int,
        module: OutputModuleName,
        *,
        name: str | None = None,
    ) -> OutputBlock: ...

    @overload
    def slot(
        self,
        number: int,
        module: ComboModuleName,
        *,
        name: str | None = None,
    ) -> tuple[InputBlock, OutputBlock]: ...

    @overload
    def slot(
        self,
        number: int,
        module: str,
        *,
        name: str | None = None,
    ) -> SlotValue: ...

    def slot(
        self,
        number: int,
        module: str,
        *,
        name: str | None = None,
    ) -> SlotValue:
        """Configure a module in the given slot and return its block(s).

        Args:
            number: Slot number (1–15 inclusive).
            module: Module part number (e.g. ``"P1-08SIM"``).  Must exist
                in :data:`~pyrung.circuitpy.catalog.MODULE_CATALOG`.
            name: Optional custom name prefix for tags in this slot.
                Defaults to ``"Slot{number}"``.

        Returns:
            - :class:`~pyrung.core.memory_block.InputBlock` for input-only modules.
            - :class:`~pyrung.core.memory_block.OutputBlock` for output-only modules.
            - ``tuple[InputBlock, OutputBlock]`` for combo modules.

        Raises:
            ValueError: If *number* is out of range, *module* is not in the
                catalog, or the slot is already configured.
        """
        if not isinstance(number, int) or number < 1 or number > MAX_SLOTS:
            msg = f"Slot number must be 1–{MAX_SLOTS}, got {number!r}."
            raise ValueError(msg)

        if number in self._slots:
            existing = self._slots[number][0]
            msg = f"Slot {number} is already configured with {existing.part_number}."
            raise ValueError(msg)

        spec = MODULE_CATALOG.get(module)
        if spec is None:
            msg = (
                f"Unknown module {module!r}. "
                f"See pyrung.circuitpy.catalog.MODULE_CATALOG for valid part numbers."
            )
            raise ValueError(msg)

        prefix = name if name is not None else f"Slot{number}"
        result = self._build_blocks(prefix, spec)
        self._slots[number] = (spec, result)
        return result

    @property
    def configured_slots(self) -> dict[int, ModuleSpec]:
        """Mapping of slot number → :class:`ModuleSpec` for all configured slots."""
        return {num: spec for num, (spec, _) in self._slots.items()}

    def get_slot(self, number: int) -> SlotValue:
        """Retrieve the block(s) for an already-configured slot.

        Raises:
            ValueError: If the slot has not been configured.
        """
        if number not in self._slots:
            msg = f"Slot {number} is not configured."
            raise ValueError(msg)
        return self._slots[number][1]

    # ------------------------------------------------------------------
    # internals
    # ------------------------------------------------------------------

    @staticmethod
    def _build_blocks(
        prefix: str,
        spec: ModuleSpec,
    ) -> SlotValue:
        input_group = spec.input_group
        output_group = spec.output_group

        inputs: InputBlock | None = None
        outputs: OutputBlock | None = None

        if input_group is not None:
            input_name = f"{prefix}_In" if spec.is_combo else prefix
            inputs = InputBlock(
                name=input_name,
                type=input_group.tag_type,
                start=1,
                end=input_group.count,
                address_formatter=_make_formatter(input_name),
            )

        if output_group is not None:
            output_name = f"{prefix}_Out" if spec.is_combo else prefix
            outputs = OutputBlock(
                name=output_name,
                type=output_group.tag_type,
                start=1,
                end=output_group.count,
                address_formatter=_make_formatter(output_name),
            )

        if inputs is not None and outputs is not None:
            return (inputs, outputs)
        if inputs is not None:
            return inputs
        # output_group is guaranteed non-None if input_group is None
        # (every module in the catalog has at least one group).
        assert outputs is not None  # noqa: S101
        return outputs

    def __repr__(self) -> str:
        if not self._slots:
            return "P1AM()"
        configured = ", ".join(
            f"{n}={spec.part_number}" for n, (spec, _) in sorted(self._slots.items())
        )
        return f"P1AM({configured})"



"""Patch/force override state for PLCRunner."""

from __future__ import annotations

from collections.abc import Callable, Mapping
from contextlib import contextmanager
from types import MappingProxyType
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any

    from pyrung.core.context import ScanContext
    from pyrung.core.tag import Tag


class InputOverrideManager:
    """Encapsulates one-shot patches and persistent force overrides."""

    def __init__(self, *, is_read_only: Callable[[str], bool]) -> None:
        self._is_read_only = is_read_only
        self._pending_patches: dict[str, bool | int | float | str] = {}
        self._forces: dict[str, bool | int | float | str] = {}

    @property
    def pending_patches(self) -> dict[str, bool | int | float | str]:
        return self._pending_patches

    @property
    def forces_mutable(self) -> dict[str, bool | int | float | str]:
        return self._forces

    def _normalize_tag_name(self, tag: str | Tag, *, method: str) -> str:
        from pyrung.core.tag import Tag as TagClass

        if isinstance(tag, TagClass):
            return tag.name
        if isinstance(tag, str):
            return tag
        raise TypeError(f"{method}() keys must be str or Tag, got {type(tag).__name__}")

    def _normalize_tag_updates(
        self,
        tags: Mapping[str, bool | int | float | str]
        | Mapping[Tag, bool | int | float | str]
        | Mapping[str | Tag, bool | int | float | str],
        *,
        method: str,
    ) -> dict[str, bool | int | float | str]:
        normalized: dict[str, bool | int | float | str] = {}
        for key, value in tags.items():
            name = self._normalize_tag_name(key, method=method)
            if self._is_read_only(name):
                raise ValueError(f"Tag '{name}' is read-only system point and cannot be written")
            normalized[name] = value
        return normalized

    def patch(
        self,
        tags: Mapping[str, bool | int | float | str]
        | Mapping[Tag, bool | int | float | str]
        | Mapping[str | Tag, bool | int | float | str],
    ) -> None:
        self._pending_patches.update(self._normalize_tag_updates(tags, method="patch"))

    def add_force(self, tag: str | Tag, value: bool | int | float | str) -> None:
        name = self._normalize_tag_name(tag, method="add_force")
        if self._is_read_only(name):
            raise ValueError(f"Tag '{name}' is read-only system point and cannot be written")
        self._forces[name] = value

    def remove_force(self, tag: str | Tag) -> None:
        name = self._normalize_tag_name(tag, method="remove_force")
        if name not in self._forces:
            raise KeyError(name)
        del self._forces[name]

    def clear_forces(self) -> None:
        self._forces.clear()

    @contextmanager
    def force(
        self,
        overrides: Mapping[str, bool | int | float | str]
        | Mapping[Tag, bool | int | float | str]
        | Mapping[str | Tag, bool | int | float | str],
    ) -> Iterator[None]:
        snapshot = self._forces.copy()
        try:
            for tag, value in overrides.items():
                self.add_force(tag, value)
            yield None
        finally:
            self._forces.clear()
            self._forces.update(snapshot)

    @property
    def forces(self) -> Mapping[str, bool | int | float | str]:
        return MappingProxyType(self._forces)

    def get_live_override(self, name: str) -> tuple[bool, Any]:
        if name in self._pending_patches:
            return True, self._pending_patches[name]
        if name in self._forces:
            return True, self._forces[name]
        return False, None

    def apply_pre_scan(self, ctx: ScanContext) -> None:
        if self._pending_patches:
            ctx.set_tags(self._pending_patches)
            self._pending_patches.clear()

        if self._forces:
            ctx.set_tags(self._forces)

    def apply_post_logic(self, ctx: ScanContext) -> None:
        if self._forces:
            ctx.set_tags(self._forces)



"""Strict prevalidation helpers for Click ladder export."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, NoReturn

from pyrung.core.rung import Rung

from .types import Issue, LadderExportError

if TYPE_CHECKING:
    from pyrung.click.tag_map import TagMap
    from pyrung.core.program import Program


# ---- Strict prevalidation mixin ----
class _ValidationMixin:
    """Run strict checks before lowering ladder logic into Click CSV rows."""

    _tag_map: TagMap
    _program: Program

    if TYPE_CHECKING:

        def _raise_issue(self, *, path: str, message: str, source: Any) -> NoReturn: ...

    def _run_precheck(self) -> None:
        report = self._tag_map.validate(self._program, mode="strict")
        # We intentionally fail on warnings and hints to keep export deterministic.
        findings = [*report.errors, *report.warnings, *report.hints]
        if findings:
            issues: list[Issue] = []
            for finding in findings:
                issues.append(
                    {
                        "path": finding.location,
                        "message": f"{finding.code}: {finding.message}",
                        "source_file": None,
                        "source_line": None,
                    }
                )
            raise LadderExportError(issues)

        # Click ladder v1 does not support nested call instructions.
        for subroutine_name in sorted(self._program.subroutines):
            for rung_index, rung in enumerate(self._program.subroutines[subroutine_name]):
                self._assert_no_nested_subroutine_calls(
                    rung,
                    path=f"subroutine[{subroutine_name}].rung[{rung_index}]",
                )

    def _assert_no_nested_subroutine_calls(self, rung: Rung, *, path: str) -> None:
        for instruction_index, instruction in enumerate(rung._instructions):
            instruction_path = (
                f"{path}.instruction[{instruction_index}]({type(instruction).__name__})"
            )
            self._assert_no_nested_subroutine_calls_in_instruction(
                instruction,
                path=instruction_path,
            )

        for branch_index, branch in enumerate(rung._branches):
            self._assert_no_nested_subroutine_calls(
                branch,
                path=f"{path}.branch[{branch_index}]",
            )

    def _assert_no_nested_subroutine_calls_in_instruction(
        self,
        instruction: Any,
        *,
        path: str,
    ) -> None:
        if type(instruction).__name__ == "CallInstruction":
            self._raise_issue(
                path=path,
                message="Nested subroutine calls are not supported for Click export.",
                source=instruction,
            )

        if type(instruction).__name__ != "ForLoopInstruction":
            return

        for child_index, child_instruction in enumerate(getattr(instruction, "instructions", ())):
            child_path = f"{path}.instruction[{child_index}]({type(child_instruction).__name__})"
            self._assert_no_nested_subroutine_calls_in_instruction(
                child_instruction,
                path=child_path,
            )


__all__ = ["_ValidationMixin"]



"""NOP instruction — explicit no-operation for Click ladder programs.

Click PLCs use NOP to mark comment-only rungs in the ladder CSV.
The PLC runner treats it as a no-op; the Click ladder exporter emits
``NOP`` in the AF column.

Usage::

    from pyrung.click import nop

    comment("Section header comment")
    with Rung():
        nop()
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pyrung.core._source import _capture_source
from pyrung.core.instruction import Instruction
from pyrung.core.program.context import _require_rung_context

if TYPE_CHECKING:
    from pyrung.core.context import ScanContext


class NopInstruction(Instruction):
    """Explicit no-operation instruction for Click ladder programs.

    Occupies a rung slot so comment-only rungs survive round-trip
    through ladder CSV export/import.  No-op in the PLC runner.
    """

    INERT_WHEN_DISABLED = True

    def execute(self, ctx: ScanContext, enabled: bool) -> None:
        """No-op — NOP instructions have no runtime behaviour."""

    def is_terminal(self) -> bool:
        """NOP must be the only instruction in its rung."""
        return True


def nop() -> None:
    """No-operation instruction (NOP) for Click ladder programs.

    Explicit placeholder for comment-only rungs.  Only one ``nop()``
    is allowed per rung, and it must be the sole instruction.

    Example::

        from pyrung.click import nop

        comment("Section header")
        with Rung():
            nop()
    """
    ctx = _require_rung_context("nop")
    ctx._assert_no_pending_required_builder("nop")
    if ctx._rung._instructions:
        raise RuntimeError("nop() must be the only instruction in a rung")
    source_file, source_line = _capture_source(depth=2)
    instruction = NopInstruction()
    instruction.source_file, instruction.source_line = source_file, source_line
    ctx._rung.add_instruction(instruction)



"""Formatting helpers for DAP stack frames and trace payloads."""

from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from typing import Any

from pyrung.core.debug_trace import TraceEvent
from pyrung.core.rung import Rung
from pyrung.core.runner import ScanStep


class DAPFormatter:
    """Build DAP-facing payloads from runner/trace models."""

    def build_current_stack_frames(
        self,
        *,
        current_step: ScanStep,
        rungs: list[Rung],
        subroutine_source_map: dict[str, tuple[str, int, int | None]],
        canonical_path: Callable[[str | None], str | None],
    ) -> list[dict[str, Any]]:
        step_name = f"Rung {current_step.rung_index}"
        if current_step.kind == "branch":
            step_name = f"Branch (rung {current_step.rung_index})"
        elif current_step.kind == "subroutine":
            sub_name = current_step.subroutine_name or "subroutine"
            step_name = f"{sub_name} (rung {current_step.rung_index})"
        elif current_step.kind == "instruction":
            kind_name = current_step.instruction_kind or "Instruction"
            if current_step.subroutine_name:
                step_name = (
                    f"{kind_name} ({current_step.subroutine_name}, rung {current_step.rung_index})"
                )
            else:
                step_name = f"{kind_name} (rung {current_step.rung_index})"

        frames: list[dict[str, Any]] = [
            self._stack_frame_from_step(
                frame_id=0,
                name=step_name,
                step=current_step,
            )
        ]

        next_frame_id = 1
        for depth, subroutine_name in enumerate(reversed(current_step.call_stack)):
            innermost = depth == 0 and current_step.subroutine_name == subroutine_name
            frames.append(
                self._stack_frame_from_subroutine(
                    frame_id=next_frame_id,
                    subroutine_name=subroutine_name,
                    current_step=current_step,
                    innermost=innermost,
                    subroutine_source_map=subroutine_source_map,
                    canonical_path=canonical_path,
                )
            )
            next_frame_id += 1

        if current_step.kind != "rung" and 0 <= current_step.rung_index < len(rungs):
            frames.append(
                self.stack_frame_from_rung(
                    frame_id=next_frame_id,
                    name=f"Rung {current_step.rung_index}",
                    rung=rungs[current_step.rung_index],
                )
            )

        return frames

    def stack_frame_from_rung(
        self,
        *,
        frame_id: int,
        name: str,
        rung: Rung,
    ) -> dict[str, Any]:
        frame: dict[str, Any] = {
            "id": frame_id,
            "name": name,
            "line": int(rung.source_line or 1),
            "column": 1,
        }
        if rung.end_line is not None:
            frame["endLine"] = int(rung.end_line)
        if rung.source_file:
            source_path = str(Path(rung.source_file))
            frame["source"] = {"name": Path(source_path).name, "path": source_path}
        return frame

    def current_trace_body(
        self,
        *,
        event_result: tuple[int, int, Any] | None,
        current_scan_id: int,
        trace_version: int,
        canonical_path: Callable[[str | None], str | None],
        format_value: Callable[[Any], str],
    ) -> dict[str, Any] | None:
        if event_result is None:
            return None

        scan_id, rung_index, event = event_result
        trace_source = "live" if scan_id > current_scan_id else "inspect"
        trace = event.trace if isinstance(event.trace, TraceEvent) else None
        step_kind: str | None = event.kind
        instruction_kind: str | None = event.instruction_kind
        enabled_state: str | None = event.enabled_state
        subroutine_name: str | None = event.subroutine_name
        call_stack: list[str] = list(event.call_stack)
        source_line: int | None = event.source_line
        end_line: int | None = event.end_line if event.end_line is not None else event.source_line
        step_source_file = event.source_file

        regions = self._regions_from_trace_event(
            trace,
            canonical_path=canonical_path,
            format_value=format_value,
        )
        step_source = None
        step_source_path = canonical_path(step_source_file)
        if step_source_path:
            step_source = {"name": Path(step_source_path).name, "path": step_source_path}

        display_status = self._step_display_status_from_fields(enabled_state=enabled_state)
        display_text = self._step_display_text_from_fields(
            kind=step_kind,
            instruction_kind=instruction_kind,
            display_status=display_status,
        )

        return {
            "traceVersion": trace_version,
            "traceSource": trace_source,
            "scanId": scan_id,
            "rungId": rung_index,
            "step": {
                "kind": step_kind,
                "instructionKind": instruction_kind,
                "enabledState": enabled_state,
                "displayStatus": display_status,
                "displayText": display_text,
                "source": step_source,
                "line": source_line,
                "endLine": end_line if end_line is not None else source_line,
                "subroutineName": subroutine_name,
                "callStack": call_stack,
                "rungIndex": rung_index,
            },
            "regions": regions,
        }

    def _stack_frame_from_step(
        self,
        *,
        frame_id: int,
        name: str,
        step: ScanStep,
    ) -> dict[str, Any]:
        source_line = int(step.source_line or step.rung.source_line or 1)
        frame: dict[str, Any] = {
            "id": frame_id,
            "name": name,
            "line": source_line,
            "column": 1,
        }
        end_line = step.end_line or step.source_line
        if end_line is not None:
            frame["endLine"] = int(end_line)
        source_file = step.source_file or step.rung.source_file
        if source_file:
            source_path = str(Path(source_file))
            frame["source"] = {"name": Path(source_path).name, "path": source_path}
        return frame

    def _stack_frame_from_subroutine(
        self,
        *,
        frame_id: int,
        subroutine_name: str,
        current_step: ScanStep,
        innermost: bool,
        subroutine_source_map: dict[str, tuple[str, int, int | None]],
        canonical_path: Callable[[str | None], str | None],
    ) -> dict[str, Any]:
        source_location: tuple[str, int, int | None] | None = None
        if innermost:
            source_location = self._subroutine_source_from_step_rung(
                current_step,
                canonical_path=canonical_path,
            )
        if source_location is None:
            source_location = subroutine_source_map.get(subroutine_name)

        frame: dict[str, Any] = {
            "id": frame_id,
            "name": f"Subroutine {subroutine_name}",
            "line": 1,
            "column": 1,
        }
        if source_location is None:
            return frame

        source_path, source_line, end_line = source_location
        frame["line"] = int(source_line)
        if end_line is not None:
            frame["endLine"] = int(end_line)
        frame["source"] = {"name": Path(source_path).name, "path": source_path}
        return frame

    def _subroutine_source_from_step_rung(
        self,
        step: ScanStep,
        *,
        canonical_path: Callable[[str | None], str | None],
    ) -> tuple[str, int, int | None] | None:
        source_path = canonical_path(step.rung.source_file)
        if source_path is None or step.rung.source_line is None:
            return None
        end_line = int(step.rung.end_line) if step.rung.end_line is not None else None
        return source_path, int(step.rung.source_line), end_line

    def _regions_from_trace_event(
        self,
        trace: TraceEvent | None,
        *,
        canonical_path: Callable[[str | None], str | None],
        format_value: Callable[[Any], str],
    ) -> list[dict[str, Any]]:
        regions: list[dict[str, Any]] = []
        if not isinstance(trace, TraceEvent):
            return regions

        for region in trace.regions:
            source_body = None
            source_path = (
                canonical_path(region.source.source_file)
                if isinstance(region.source.source_file, str)
                else None
            )
            if source_path:
                source_body = {"name": Path(source_path).name, "path": source_path}

            conditions: list[dict[str, Any]] = []
            for cond in region.conditions:
                cond_source = None
                cond_path = (
                    canonical_path(cond.source_file) if isinstance(cond.source_file, str) else None
                )
                if cond_path:
                    cond_source = {"name": Path(cond_path).name, "path": cond_path}
                details = [
                    {
                        "name": str(detail.get("name", "")),
                        "value": format_value(detail.get("value")),
                    }
                    for detail in cond.details
                    if isinstance(detail, dict)
                ]
                conditions.append(
                    {
                        "source": cond_source,
                        "line": cond.source_line,
                        "expression": cond.expression,
                        "status": cond.status,
                        "value": cond.value,
                        "details": details,
                        "summary": cond.summary,
                        "annotation": cond.annotation,
                    }
                )

            regions.append(
                {
                    "kind": region.kind,
                    "enabledState": region.enabled_state,
                    "source": source_body,
                    "line": region.source.source_line,
                    "endLine": region.source.end_line,
                    "conditions": conditions,
                }
            )

        return regions

    def _step_display_status_from_fields(self, *, enabled_state: str | None) -> str:
        if enabled_state == "enabled":
            return "enabled"
        if enabled_state == "disabled_parent":
            return "skipped"
        return "disabled"

    def _step_display_text_from_fields(
        self,
        *,
        kind: str | None,
        instruction_kind: str | None,
        display_status: str,
    ) -> str:
        if display_status == "enabled":
            prefix = "[RUN]" if kind == "instruction" else "[ON]"
        elif display_status == "skipped":
            prefix = "[SKIP]"
        else:
            prefix = "[OFF]"

        if kind == "instruction":
            label = instruction_kind or "Instruction"
        elif kind == "branch":
            label = "Branch"
        elif kind == "subroutine":
            label = "Subroutine"
        else:
            label = "Rung"
        return f"{prefix} {label}"



"""Time modes for PLC simulation."""

from enum import Enum


class TimeMode(Enum):
    """Simulation time modes.

    FIXED_STEP: Each scan advances by a fixed dt, regardless of wall clock.
                Use for unit tests and deterministic simulations.

    REALTIME: Simulation clock tracks actual elapsed time.
              Use for integration tests and hardware-in-loop.
    """

    FIXED_STEP = "fixed_step"
    REALTIME = "realtime"


class TimeUnit(Enum):
    """Timer time units for Click PLC.

    The accumulator stores integer values in the specified unit.
    Conversion from dt (seconds) uses appropriate scaling.
    """

    Tms = "milliseconds"  # 1 unit = 1 ms, multiply dt by 1000
    Ts = "seconds"  # 1 unit = 1 second
    Tm = "minutes"  # 1 unit = 1 minute, divide dt by 60
    Th = "hours"  # 1 unit = 1 hour, divide dt by 3600
    Td = "days"  # 1 unit = 1 day, divide dt by 86400

    def dt_to_units(self, dt_seconds: float) -> float:
        """Convert dt in seconds to timer units (with fractional part)."""
        match self:
            case TimeUnit.Tms:
                return dt_seconds * 1000
            case TimeUnit.Ts:
                return dt_seconds
            case TimeUnit.Tm:
                return dt_seconds / 60
            case TimeUnit.Th:
                return dt_seconds / 3600
            case TimeUnit.Td:
                return dt_seconds / 86400



from __future__ import annotations

import ast
import inspect
import linecache
import textwrap
import warnings
from collections.abc import Callable
from types import FrameType
from typing import Any


class ForbiddenControlFlowError(RuntimeError):
    """Raised when Python control flow is used inside strict DSL scope."""


_IF_HINT = "Use `Rung(condition)` to express conditional logic"


_BOOL_HINT = "Use `all_of()` / `any_of()` for compound conditions"


_NOT_HINT = "Use `~Tag` for normally-closed contacts"


_LOOP_HINT = "Each rung is independent; express repeated patterns as separate rungs"


_ASSIGN_HINT = "DSL instructions write to tags directly; no intermediate Python variables needed"


_TRY_HINT = "Errors in DSL scope are programming mistakes; no recovery logic in ladder logic"


_COMPREHENSION_HINT = (
    "Build tag collections outside the Program scope, then reference them in rungs"
)


_SCOPE_HINT = "DSL scope should not mutate external Python state"


_RETURN_HINT = "Use `return_early()` for early subroutine exit; no Python control flow in DSL scope"


_IMPORT_HINT = "Move imports outside the Program/subroutine scope"


_ASSERT_HINT = "Not valid in ladder logic; handle validation outside DSL scope"


_DEF_HINT = "Define functions and classes outside the Program/subroutine scope"


_GENERIC_STMT_HINT = "Only `with ...:`, bare function calls, and `pass` are allowed in DSL scope"


DialectValidator = Callable[..., Any]


_EXPR_STMT_HINT = "Only bare call expressions are allowed as statements"


_FORBIDDEN_NODE_RULES: tuple[tuple[tuple[type[ast.AST], ...], str | None, str], ...] = (
    ((ast.If, ast.IfExp), "if/elif/else", _IF_HINT),
    ((ast.BoolOp,), "and/or", _BOOL_HINT),
    ((ast.For, ast.AsyncFor, ast.While), "for/while", _LOOP_HINT),
    ((ast.Assign, ast.AnnAssign, ast.AugAssign, ast.NamedExpr), "assignment", _ASSIGN_HINT),
    ((ast.Try,), "try/except", _TRY_HINT),
    (
        (ast.ListComp, ast.SetComp, ast.DictComp, ast.GeneratorExp, ast.comprehension),
        "comprehension/generator",
        _COMPREHENSION_HINT,
    ),
    ((ast.Global, ast.Nonlocal), "global/nonlocal", _SCOPE_HINT),
    ((ast.Yield, ast.YieldFrom, ast.Await, ast.Return), "yield/await/return", _RETURN_HINT),
    ((ast.Import, ast.ImportFrom), "import", _IMPORT_HINT),
    ((ast.Assert, ast.Raise, ast.Delete), "assert/raise/del", _ASSERT_HINT),
    ((ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef), None, _DEF_HINT),
    ((ast.Expr,), "expression statement", _EXPR_STMT_HINT),
    ((ast.stmt,), None, _GENERIC_STMT_HINT),
)


def _warn_check_skipped(target: str, reason: Exception) -> None:
    """Warn and skip strict checking when source inspection/parsing is unavailable."""
    warnings.warn(
        f"Unable to perform strict DSL control-flow check for {target}: {reason}",
        RuntimeWarning,
        stacklevel=3,
    )


def _absolute_line(node: ast.AST, line_offset: int) -> int:
    lineno = getattr(node, "lineno", 1)
    return line_offset + lineno - 1


def _describe_forbidden_node(node: ast.AST) -> tuple[str, str]:
    """Return user-facing construct label and DSL hint for a forbidden node."""
    if isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not):
        return "not", _NOT_HINT
    for node_types, label, hint in _FORBIDDEN_NODE_RULES:
        if isinstance(node, node_types):
            return (type(node).__name__ if label is None else label), hint
    return type(node).__name__, "This construct is not allowed in strict DSL scope"


def _raise_forbidden_node(
    node: ast.AST,
    *,
    filename: str,
    line_offset: int,
    opt_out_hint: str,
) -> None:
    construct, hint = _describe_forbidden_node(node)
    line = _absolute_line(node, line_offset)
    raise ForbiddenControlFlowError(
        f"{filename}:{line}: forbidden Python construct '{construct}' in strict DSL scope. "
        f"{hint}. Opt out with {opt_out_hint}."
    )


def _iter_expression_nodes(node: ast.AST) -> list[ast.AST]:
    """Iterate expression graph for a statement, excluding nested statements."""
    nodes: list[ast.AST] = []
    stack: list[ast.AST] = [node]
    while stack:
        current = stack.pop()
        nodes.append(current)
        children = list(ast.iter_child_nodes(current))
        for child in reversed(children):
            if isinstance(child, ast.stmt):
                continue
            stack.append(child)
    return nodes


def _check_expression_tree(
    node: ast.AST,
    *,
    filename: str,
    line_offset: int,
    opt_out_hint: str,
) -> None:
    for child in _iter_expression_nodes(node):
        if isinstance(child, ast.BoolOp):
            _raise_forbidden_node(
                child,
                filename=filename,
                line_offset=line_offset,
                opt_out_hint=opt_out_hint,
            )
        if isinstance(child, ast.UnaryOp) and isinstance(child.op, ast.Not):
            _raise_forbidden_node(
                child,
                filename=filename,
                line_offset=line_offset,
                opt_out_hint=opt_out_hint,
            )
        if isinstance(
            child,
            (
                ast.IfExp,
                ast.NamedExpr,
                ast.Await,
                ast.Yield,
                ast.YieldFrom,
                ast.ListComp,
                ast.SetComp,
                ast.DictComp,
                ast.GeneratorExp,
                ast.comprehension,
            ),
        ):
            _raise_forbidden_node(
                child,
                filename=filename,
                line_offset=line_offset,
                opt_out_hint=opt_out_hint,
            )


def _check_statement_list(
    statements: list[ast.stmt],
    *,
    filename: str,
    line_offset: int,
    opt_out_hint: str,
) -> None:
    for statement in statements:
        if isinstance(statement, ast.Pass):
            continue

        if isinstance(statement, ast.Expr):
            if not isinstance(statement.value, ast.Call):
                _raise_forbidden_node(
                    statement,
                    filename=filename,
                    line_offset=line_offset,
                    opt_out_hint=opt_out_hint,
                )
            _check_expression_tree(
                statement.value,
                filename=filename,
                line_offset=line_offset,
                opt_out_hint=opt_out_hint,
            )
            continue

        if isinstance(statement, ast.With):
            for item in statement.items:
                _check_expression_tree(
                    item.context_expr,
                    filename=filename,
                    line_offset=line_offset,
                    opt_out_hint=opt_out_hint,
                )
                if item.optional_vars is not None:
                    _check_expression_tree(
                        item.optional_vars,
                        filename=filename,
                        line_offset=line_offset,
                        opt_out_hint=opt_out_hint,
                    )
            _check_statement_list(
                statement.body,
                filename=filename,
                line_offset=line_offset,
                opt_out_hint=opt_out_hint,
            )
            continue

        _raise_forbidden_node(
            statement,
            filename=filename,
            line_offset=line_offset,
            opt_out_hint=opt_out_hint,
        )


def _check_function_body_strict(
    fn: Callable[[], None],
    *,
    opt_out_hint: str,
    source_label: str,
) -> None:
    try:
        source_lines, start_line = inspect.getsourcelines(fn)
        source = textwrap.dedent("".join(source_lines))
        module = ast.parse(source)
    except (OSError, TypeError, SyntaxError) as exc:
        _warn_check_skipped(source_label, exc)
        return

    function_nodes = [
        node for node in module.body if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
    ]
    if not function_nodes:
        _warn_check_skipped(source_label, RuntimeError("function body AST not found"))
        return

    filename = inspect.getsourcefile(fn) or inspect.getfile(fn)
    _check_statement_list(
        function_nodes[0].body,
        filename=filename,
        line_offset=start_line - 1,
        opt_out_hint=opt_out_hint,
    )


def _find_enclosing_with(module: ast.Module, line_number: int) -> ast.With | None:
    matches: list[ast.With] = []
    for node in ast.walk(module):
        if not isinstance(node, ast.With):
            continue
        end_line = getattr(node, "end_lineno", node.lineno)
        if node.lineno <= line_number <= end_line:
            matches.append(node)

    if not matches:
        return None
    return max(matches, key=lambda node: node.lineno)


def _check_with_body_from_frame(frame: FrameType, *, opt_out_hint: str) -> None:
    code = frame.f_code
    source_label = f"{code.co_filename}:{frame.f_lineno}"
    try:
        if code.co_name == "<module>":
            # getsourcelines on module-level code objects only returns the
            # first statement (Python's getblock stops early), so read the
            # full file instead.
            filename = inspect.getsourcefile(code) or code.co_filename
            source_lines = linecache.getlines(filename, frame.f_globals)
            start_line = 1
        else:
            source_lines, start_line = inspect.getsourcelines(code)
        source = textwrap.dedent("".join(source_lines))
        module = ast.parse(source)
    except (OSError, TypeError, SyntaxError) as exc:
        _warn_check_skipped(source_label, exc)
        return

    relative_line = frame.f_lineno - start_line + 1
    with_node = _find_enclosing_with(module, relative_line)
    if with_node is None:
        _warn_check_skipped(source_label, RuntimeError("enclosing with-statement AST not found"))
        return

    filename = inspect.getsourcefile(code) or code.co_filename
    _check_statement_list(
        with_node.body,
        filename=filename,
        line_offset=start_line - 1,
        opt_out_hint=opt_out_hint,
    )



"""Tests for multi-file project output (``ladder_to_pyrung_project``)."""

from __future__ import annotations

import importlib
import sys
from pathlib import Path

from pyrung.click import (
    TagMap,
    c,
    ct,
    ctd,
    ladder_to_pyrung_project,
    pyrung_to_ladder,
    t,
    td,
    x,
    y,
)
from pyrung.core import (
    Block,
    Bool,
    Dint,
    Int,
    Program,
    Rung,
    TagType,
    Tms,
    any_of,
)
from pyrung.core.program import (
    call,
    count_up,
    on_delay,
    out,
    reset,
    subroutine,
)

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _project_from_program(
    program: Program,
    tag_map: TagMap,
    tmp_path: Path,
    *,
    nicknames: dict[str, str] | None = None,
) -> dict[str, str]:
    """Program -> CSV -> ladder_to_pyrung_project -> files dict."""
    bundle = pyrung_to_ladder(program, tag_map)
    csv_dir = tmp_path / "csv"
    bundle.write(csv_dir)
    return ladder_to_pyrung_project(csv_dir, nicknames=nicknames)


def _exec_project(files: dict[str, str], tmp_path: Path) -> dict:
    """Write project files to disk and import them, returning main module namespace."""
    project_dir = tmp_path / "project"
    for rel_path, content in files.items():
        fpath = project_dir / rel_path
        fpath.parent.mkdir(parents=True, exist_ok=True)
        fpath.write_text(content, encoding="utf-8")

    old_path = sys.path[:]
    sys.path.insert(0, str(project_dir))
    # Track modules to clean up
    added_modules: list[str] = []
    try:
        tags_mod = importlib.import_module("tags")
        added_modules.append("tags")

        # Import subroutine modules
        for rel_path in sorted(files):
            if (
                rel_path.startswith("subroutines/")
                and rel_path.endswith(".py")
                and rel_path != "subroutines/__init__.py"
            ):
                mod_name = rel_path[:-3].replace("/", ".")
                importlib.import_module(mod_name)
                added_modules.append(mod_name)

        # subroutines package itself
        if "subroutines/__init__.py" in files:
            if "subroutines" not in sys.modules:
                importlib.import_module("subroutines")
            added_modules.append("subroutines")

        main_mod = importlib.import_module("main")
        added_modules.append("main")

        return {
            "logic": main_mod.logic,
            "mapping": tags_mod.mapping,
        }
    finally:
        sys.path[:] = old_path
        for key in added_modules:
            sys.modules.pop(key, None)


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------


class TestProjectBasic:
    """Basic project generation (no subroutines)."""

    def test_basic_files(self, tmp_path: Path):
        """Program with no subroutines produces main.py and tags.py only."""
        Button = Bool("Button")
        Light = Bool("Light")

        with Program() as logic:
            with Rung(Button):
                out(Light)

        mapping = TagMap({Button: x[1], Light: y[1]}, include_system=False)
        files = _project_from_program(logic, mapping, tmp_path)

        assert "main.py" in files
        assert "tags.py" in files
        assert "subroutines/__init__.py" not in files

    def test_tags_file_has_declarations_and_mapping(self, tmp_path: Path):
        """tags.py contains tag declarations and TagMap."""
        Button = Bool("Button")
        Light = Bool("Light")

        with Program() as logic:
            with Rung(Button):
                out(Light)

        mapping = TagMap({Button: x[1], Light: y[1]}, include_system=False)
        files = _project_from_program(logic, mapping, tmp_path)

        tags_py = files["tags.py"]
        assert "Generated by pyrung" in tags_py
        assert "Bool" in tags_py
        assert "TagMap" in tags_py
        assert "mapping = TagMap(" in tags_py
        assert "X001" in tags_py or "Button" in tags_py

    def test_main_file_has_program(self, tmp_path: Path):
        """main.py contains Program context and imports from tags."""
        Button = Bool("Button")
        Light = Bool("Light")

        with Program() as logic:
            with Rung(Button):
                out(Light)

        mapping = TagMap({Button: x[1], Light: y[1]}, include_system=False)
        files = _project_from_program(logic, mapping, tmp_path)

        main_py = files["main.py"]
        assert "Generated by pyrung" in main_py
        assert "from pyrung import" in main_py
        assert "from tags import" in main_py
        assert "with Program() as logic:" in main_py

    def test_tags_no_inline_comments(self, tmp_path: Path):
        """tags.py suppresses inline address comments."""
        Button = Bool("Button")
        Light = Bool("Light")

        with Program() as logic:
            with Rung(Button):
                out(Light)

        mapping = TagMap({Button: x[1], Light: y[1]}, include_system=False)
        # Use nicknames so that comments would normally appear
        files = (
            ladder_to_pyrung_project(
                tmp_path / "csv",  # doesn't exist yet, use direct bundle
                nicknames={"X001": "Button", "Y001": "Light"},
            )
            if False
            else _project_from_program(
                logic, mapping, tmp_path, nicknames={"X001": "Button", "Y001": "Light"}
            )
        )

        tags_py = files["tags.py"]
        assert "# X001" not in tags_py
        assert "# Y001" not in tags_py
        # But nicknames should still be used as variable names
        assert 'Button = Bool("Button")' in tags_py

    def test_exec_round_trip_no_subroutines(self, tmp_path: Path):
        """Generated project files can be imported and produce valid program."""
        Button = Bool("Button")
        Light = Bool("Light")

        with Program() as logic:
            with Rung(Button):
                out(Light)

        mapping = TagMap({Button: x[1], Light: y[1]}, include_system=False)
        files = _project_from_program(logic, mapping, tmp_path)

        ns = _exec_project(files, tmp_path)
        # Re-export and verify
        bundle = pyrung_to_ladder(ns["logic"], ns["mapping"])
        assert len(bundle.main_rows) > 0

    def test_plain_block_range_project_imports_block_symbol(self, tmp_path: Path):
        """Project output should import reconstructed plain blocks, not range helpers."""
        import pyclickplc
        from pyclickplc.addresses import AddressRecord, get_addr_key
        from pyclickplc.banks import DataType

        Enable = Bool("Enable")
        Bits = Block("Bits", TagType.BOOL, 1, 3)

        with Program() as logic:
            with Rung(Enable):
                out(Bits[1])
                reset(Bits.select(1, 3))

        mapping = TagMap({Enable: x[1], Bits: c.select(1004, 1006)}, include_system=False)
        bundle = pyrung_to_ladder(logic, mapping)
        csv_dir = tmp_path / "csv"
        bundle.write(csv_dir)

        nick_path = tmp_path / "nicknames.csv"
        pyclickplc.write_csv(
            nick_path,
            {
                get_addr_key("C", 1001): AddressRecord(
                    memory_type="C",
                    address=1001,
                    nickname="",
                    comment="<CmdTagBits:block>",
                    initial_value="0",
                    retentive=False,
                    data_type=DataType.BIT,
                ),
                get_addr_key("C", 1004): AddressRecord(
                    memory_type="C",
                    address=1004,
                    nickname="Cmd_Mode_Production",
                    comment="",
                    initial_value="0",
                    retentive=False,
                    data_type=DataType.BIT,
                ),
                get_addr_key("C", 1019): AddressRecord(
                    memory_type="C",
                    address=1019,
                    nickname="",
                    comment="</CmdTagBits:block>",
                    initial_value="0",
                    retentive=False,
                    data_type=DataType.BIT,
                ),
                get_addr_key("X", 1): AddressRecord(
                    memory_type="X",
                    address=1,
                    nickname="Enable",
                    comment="",
                    initial_value="0",
                    retentive=False,
                    data_type=DataType.BIT,
                ),
            },
        )

        files = ladder_to_pyrung_project(csv_dir, nickname_csv=nick_path)

        assert 'CmdTagBits = Block("CmdTagBits", TagType.BOOL, 1, 19)' in files["tags.py"]
        assert "Cmd_Mode_Production = CmdTagBits[4]" in files["tags.py"]
        assert "from tags import mapping" in files["main.py"]
        assert "Cmd_Mode_Production" in files["main.py"]
        assert "CmdTagBits" in files["main.py"]
        assert "out(Cmd_Mode_Production)" in files["main.py"]
        assert "reset(CmdTagBits.select(4, 6))" in files["main.py"]
        assert "C1004_to_C1006" not in files["tags.py"]
        assert "C1004_to_C1006" not in files["main.py"]

    def test_plain_block_project_lifts_uniform_retentive_to_block_default(self, tmp_path: Path):
        """Uniform plain-block retentive policy should lift to the Block constructor."""
        import pyclickplc
        from pyclickplc.addresses import AddressRecord, get_addr_key
        from pyclickplc.banks import DataType

        Enable = Bool("Enable")
        Bits = Block("Bits", TagType.BOOL, 1, 3)

        with Program() as logic:
            with Rung(Enable):
                out(Bits[1])
                reset(Bits.select(1, 3))

        mapping = TagMap({Enable: x[1], Bits: c.select(1001, 1003)}, include_system=False)
        bundle = pyrung_to_ladder(logic, mapping)
        csv_dir = tmp_path / "csv"
        bundle.write(csv_dir)

        nick_path = tmp_path / "nicknames.csv"
        pyclickplc.write_csv(
            nick_path,
            {
                get_addr_key("C", 1001): AddressRecord(
                    memory_type="C",
                    address=1001,
                    nickname="AlarmCoil_PLC",
                    comment="<AlarmCoil:block>",
                    initial_value="0",
                    retentive=True,
                    data_type=DataType.BIT,
                ),
                get_addr_key("C", 1002): AddressRecord(
                    memory_type="C",
                    address=1002,
                    nickname="AlarmCoil_LostData",
                    comment="",
                    initial_value="0",
                    retentive=True,
                    data_type=DataType.BIT,
                ),
                get_addr_key("C", 1003): AddressRecord(
                    memory_type="C",
                    address=1003,
                    nickname="AlarmCoil_Watchdog",
                    comment="</AlarmCoil:block>",
                    initial_value="0",
                    retentive=True,
                    data_type=DataType.BIT,
                ),
                get_addr_key("X", 1): AddressRecord(
                    memory_type="X",
                    address=1,
                    nickname="Enable",
                    comment="",
                    initial_value="0",
                    retentive=False,
                    data_type=DataType.BIT,
                ),
            },
        )

        files = ladder_to_pyrung_project(csv_dir, nickname_csv=nick_path)

        assert (
            'AlarmCoil = Block("AlarmCoil", TagType.BOOL, 1, 3, retentive=True)' in files["tags.py"]
        )
        assert "AlarmCoil.slot(1, name='AlarmCoil_PLC')" in files["tags.py"]
        assert "AlarmCoil.slot(2, name='AlarmCoil_LostData')" in files["tags.py"]
        assert "AlarmCoil.slot(3, name='AlarmCoil_Watchdog')" in files["tags.py"]
        assert (
            "retentive=True"
            not in files["tags.py"].split(
                'AlarmCoil = Block("AlarmCoil", TagType.BOOL, 1, 3, retentive=True)', 1
            )[1]
        )


class TestProjectWithSubroutines:
    """Project generation with subroutines."""

    def _make_program_with_sub(self):
        Button = Bool("Button")
        Light = Bool("Light")
        SubLight = Bool("SubLight")

        with Program() as logic:
            with Rung(Button):
                out(Light)
                call("init")

            with subroutine("init"):
                with Rung():
                    out(SubLight)

        mapping = TagMap(
            {Button: x[1], Light: y[1], SubLight: y[2]},
            include_system=False,
        )
        return logic, mapping

    def test_file_structure(self, tmp_path: Path):
        """Project with subroutines has correct file layout."""
        logic, mapping = self._make_program_with_sub()
        files = _project_from_program(logic, mapping, tmp_path)

        assert "main.py" in files
        assert "tags.py" in files
        assert "subroutines/__init__.py" in files
        assert "subroutines/init.py" in files
        assert files["subroutines/__init__.py"] == ""

    def test_subroutine_has_decorator(self, tmp_path: Path):
        """Subroutine file uses @subroutine decorator pattern."""
        logic, mapping = self._make_program_with_sub()
        files = _project_from_program(logic, mapping, tmp_path)

        sub_py = files["subroutines/init.py"]
        assert '@subroutine("init")' in sub_py
        assert "def init():" in sub_py
        assert "from tags import" in sub_py

    def test_main_calls_func_not_string(self, tmp_path: Path):
        """main.py uses call(func) not call('name')."""
        logic, mapping = self._make_program_with_sub()
        files = _project_from_program(logic, mapping, tmp_path)

        main_py = files["main.py"]
        assert "call(init)" in main_py
        assert 'call("init")' not in main_py
        assert "from subroutines.init import init" in main_py

    def test_exec_round_trip_with_subroutine(self, tmp_path: Path):
        """Full exec round-trip with subroutine."""
        logic, mapping = self._make_program_with_sub()

        # Get original CSV
        original_bundle = pyrung_to_ladder(logic, mapping)
        original_rows = list(original_bundle.main_rows)

        files = _project_from_program(logic, mapping, tmp_path)
        ns = _exec_project(files, tmp_path)

        # Re-export
        reproduced_bundle = pyrung_to_ladder(ns["logic"], ns["mapping"])
        reproduced_rows = list(reproduced_bundle.main_rows)

        assert original_rows == reproduced_rows

    def test_subroutine_with_conditions(self, tmp_path: Path):
        """Subroutine with conditional rungs round-trips."""
        Enable = Bool("Enable")
        Cond = Bool("Cond")
        Y1 = Bool("Y1")
        Y2 = Bool("Y2")

        with Program() as logic:
            with Rung(Enable):
                call("worker")

            with subroutine("worker"):
                with Rung(Cond):
                    out(Y1)
                with Rung():
                    out(Y2)

        mapping = TagMap(
            {Enable: x[1], Cond: x[2], Y1: y[1], Y2: y[2]},
            include_system=False,
        )

        original_bundle = pyrung_to_ladder(logic, mapping)
        original_rows = list(original_bundle.main_rows)

        files = _project_from_program(logic, mapping, tmp_path)
        ns = _exec_project(files, tmp_path)

        reproduced_bundle = pyrung_to_ladder(ns["logic"], ns["mapping"])
        reproduced_rows = list(reproduced_bundle.main_rows)

        assert original_rows == reproduced_rows


class TestPerFileImports:
    """Each file should only import what it uses."""

    def test_subroutine_imports_only_its_tags(self, tmp_path: Path):
        """Sub1 uses X001/Y001, Sub2 uses X002/Y002 — imports are isolated."""
        X1 = Bool("X1")
        X2 = Bool("X2")
        Y1 = Bool("Y1")
        Y2 = Bool("Y2")
        Main = Bool("Main")

        with Program() as logic:
            with Rung(Main):
                call("sub1")
                call("sub2")

            with subroutine("sub1"):
                with Rung(X1):
                    out(Y1)

            with subroutine("sub2"):
                with Rung(X2):
                    out(Y2)

        mapping = TagMap(
            {Main: c[1], X1: x[1], X2: x[2], Y1: y[1], Y2: y[2]},
            include_system=False,
        )
        files = _project_from_program(logic, mapping, tmp_path)

        sub1 = files["subroutines/sub1.py"]
        sub2 = files["subroutines/sub2.py"]

        # sub1 imports X1, Y1 but not X2, Y2
        assert "X001" in sub1 or "X1" in sub1
        assert "Y001" in sub1 or "Y1" in sub1
        # sub2 imports X2, Y2 but not X1, Y1
        assert "X002" in sub2 or "X2" in sub2
        assert "Y002" in sub2 or "Y2" in sub2

    def test_main_imports_mapping(self, tmp_path: Path):
        """main.py always imports mapping from tags."""
        Button = Bool("Button")
        Light = Bool("Light")

        with Program() as logic:
            with Rung(Button):
                out(Light)

        mapping = TagMap({Button: x[1], Light: y[1]}, include_system=False)
        files = _project_from_program(logic, mapping, tmp_path)

        main_py = files["main.py"]
        assert "mapping" in main_py
        # mapping should come from tags
        from_tags_line = [ln for ln in main_py.splitlines() if ln.startswith("from tags import")]
        assert len(from_tags_line) == 1
        assert "mapping" in from_tags_line[0]

    def test_main_omits_any_of_for_two_way_bool_or(self, tmp_path: Path):
        """Two-way BOOL OR renders with | and does not import any_of."""
        A = Bool("A")
        B = Bool("B")
        Y = Bool("Y")

        with Program() as logic:
            with Rung(any_of(A, B)):
                out(Y)

        mapping = TagMap({A: x[1], B: x[2], Y: y[1]}, include_system=False)
        files = _project_from_program(logic, mapping, tmp_path)

        main_py = files["main.py"]
        assert "with Rung(" in main_py
        assert " | " in main_py
        assert "any_of" not in main_py

    def test_subroutine_imports_click_block_for_raw_range(self, tmp_path: Path):
        """Subroutine using reset(c.select(...)) must import c from pyrung.click."""
        Enable = Bool("Enable")
        Coil = Bool("Coil")
        Bits = Block("Bits", TagType.BOOL, 1, 3)

        with Program() as logic:
            with Rung(Enable):
                call("worker")

            with subroutine("worker"):
                with Rung():
                    out(Coil)
                    reset(Bits.select(1, 3))

        mapping = TagMap(
            {Enable: x[1], Coil: y[1], Bits: c.select(1004, 1006)},
            include_system=False,
        )
        files = _project_from_program(logic, mapping, tmp_path)
        sub_py = files["subroutines/worker.py"]

        # Must have a pyrung.click import that includes 'c'
        click_import = [
            ln for ln in sub_py.splitlines() if ln.startswith("from pyrung.click import")
        ]
        assert click_import, "subroutine should have a pyrung.click import line"
        assert "c" in click_import[0].split("import")[1]

        # Verify the project can actually be imported and executed
        ns = _exec_project(files, tmp_path)
        assert ns["logic"] is not None

    def test_subroutine_imports_click_block_for_pointer(self, tmp_path: Path):
        """Subroutine using copy(DH[idx], dest) must import dh from pyrung.click."""
        from pyrung.click.codegen.collector import _collect_operands
        from pyrung.click.codegen.models import (
            _AnalyzedRung,
            _InstructionInfo,
            _SubroutineInfo,
        )
        from pyrung.click.codegen.project_emitter import _generate_project

        main_rung = _AnalyzedRung(
            comment=None,
            condition_tree=None,
            instructions=[_InstructionInfo(af_token='call("worker")', branch_tree=None, pins=[])],
        )
        sub_rung = _AnalyzedRung(
            comment=None,
            condition_tree=None,
            instructions=[
                _InstructionInfo(af_token="copy(DH[DS134],DH051)", branch_tree=None, pins=[]),
            ],
        )

        subroutines = [_SubroutineInfo(name="worker", analyzed=[sub_rung])]
        all_rungs = [main_rung, sub_rung]
        collection = _collect_operands(all_rungs, None)
        collection.has_subroutine = True

        files = _generate_project([main_rung], collection, None, subroutines)
        sub_py = files["subroutines/worker.py"]

        # Must have dh imported from pyrung.click
        click_import = [
            ln for ln in sub_py.splitlines() if ln.startswith("from pyrung.click import")
        ]
        assert click_import, "subroutine should have a pyrung.click import line"
        assert " dh" in click_import[0] or ",dh" in click_import[0]

        # Should use lowercase block variable, not uppercase Click prefix
        assert "dh[" in sub_py
        assert "DH[" not in sub_py


class TestSubroutineCallsSubroutine:
    """Subroutine that calls another subroutine.

    Note: Click hardware doesn't support nested subroutine calls, so
    we test the codegen import structure directly (not via CSV round-trip).
    """

    def test_cross_sub_import(self, tmp_path: Path):
        """sub_a calls sub_b — sub_a.py imports from subroutines.sub_b."""
        from pyrung.click.codegen.collector import _collect_operands
        from pyrung.click.codegen.models import (
            _AnalyzedRung,
            _InstructionInfo,
            _SubroutineInfo,
        )
        from pyrung.click.codegen.project_emitter import _generate_project

        # Build analyzed rungs by hand (bypassing CSV which rejects nested calls)
        main_rung = _AnalyzedRung(
            comment=None,
            condition_tree=None,
            instructions=[_InstructionInfo(af_token='call("sub_a")', branch_tree=None, pins=[])],
        )
        sub_a_rung = _AnalyzedRung(
            comment=None,
            condition_tree=None,
            instructions=[
                _InstructionInfo(af_token='call("sub_b")', branch_tree=None, pins=[]),
                _InstructionInfo(af_token="out(Y001)", branch_tree=None, pins=[]),
            ],
        )
        sub_b_rung = _AnalyzedRung(
            comment=None,
            condition_tree=None,
            instructions=[_InstructionInfo(af_token="out(Y002)", branch_tree=None, pins=[])],
        )

        subroutines = [
            _SubroutineInfo(name="sub_a", analyzed=[sub_a_rung]),
            _SubroutineInfo(name="sub_b", analyzed=[sub_b_rung]),
        ]

        all_rungs = [main_rung, sub_a_rung, sub_b_rung]
        collection = _collect_operands(all_rungs, None)
        collection.has_subroutine = True

        files = _generate_project([main_rung], collection, None, subroutines)

        sub_a = files["subroutines/sub_a.py"]
        assert "from subroutines.sub_b import sub_b" in sub_a
        assert "call(sub_b)" in sub_a


class TestNicknames:
    """Nickname variable names in project output."""

    def test_nicknames_in_all_files(self, tmp_path: Path):
        """Nicknames appear in tags.py and are used in main/sub files."""
        Button = Bool("Button")
        Light = Bool("Light")
        SubLight = Bool("SubLight")

        with Program() as logic:
            with Rung(Button):
                out(Light)
                call("init")

            with subroutine("init"):
                with Rung():
                    out(SubLight)

        mapping = TagMap(
            {Button: x[1], Light: y[1], SubLight: y[2]},
            include_system=False,
        )
        nicks = {"X001": "Button", "Y001": "Light", "Y002": "SubLight"}
        files = _project_from_program(logic, mapping, tmp_path, nicknames=nicks)

        tags_py = files["tags.py"]
        assert 'Button = Bool("Button")' in tags_py
        assert 'Light = Bool("Light")' in tags_py

        main_py = files["main.py"]
        assert "Button" in main_py
        assert "Light" in main_py

        sub_py = files["subroutines/init.py"]
        assert "SubLight" in sub_py


class TestDiskWrite:
    """Writing project to disk."""

    def test_output_dir(self, tmp_path: Path):
        """ladder_to_pyrung_project writes files to output_dir."""
        Button = Bool("Button")
        Light = Bool("Light")

        with Program() as logic:
            with Rung(Button):
                out(Light)

        mapping = TagMap({Button: x[1], Light: y[1]}, include_system=False)
        bundle = pyrung_to_ladder(logic, mapping)
        csv_dir = tmp_path / "csv"
        bundle.write(csv_dir)

        out_dir = tmp_path / "output"
        files = ladder_to_pyrung_project(csv_dir, output_dir=out_dir)

        assert (out_dir / "main.py").exists()
        assert (out_dir / "tags.py").exists()
        assert (out_dir / "main.py").read_text(encoding="utf-8") == files["main.py"]


class TestTimerCounterInstructions:
    """Project output with timer and counter instructions."""

    def test_timer_round_trip(self, tmp_path: Path):
        """Timer instructions produce correct project output."""
        Enable = Bool("Enable")
        Done = Bool("Done")
        Acc = Int("Acc")

        with Program() as logic:
            with Rung(Enable):
                on_delay(Done, Acc, preset=1000, unit=Tms)

        mapping = TagMap(
            {Enable: x[1], Done: t[1], Acc: td[1]},
            include_system=False,
        )
        files = _project_from_program(logic, mapping, tmp_path)
        ns = _exec_project(files, tmp_path)

        original_bundle = pyrung_to_ladder(logic, mapping)
        reproduced_bundle = pyrung_to_ladder(ns["logic"], ns["mapping"])
        assert list(original_bundle.main_rows) == list(reproduced_bundle.main_rows)

    def test_counter_round_trip(self, tmp_path: Path):
        """Counter instructions produce correct project output."""
        Enable = Bool("Enable")
        ResetCond = Bool("ResetCond")
        Done = Bool("Done")
        Acc = Dint("Acc")

        with Program() as logic:
            with Rung(Enable):
                count_up(Done, Acc, preset=10).reset(ResetCond)

        mapping = TagMap(
            {Enable: x[1], ResetCond: x[2], Done: ct[1], Acc: ctd[1]},
            include_system=False,
        )
        files = _project_from_program(logic, mapping, tmp_path)
        ns = _exec_project(files, tmp_path)

        original_bundle = pyrung_to_ladder(logic, mapping)
        reproduced_bundle = pyrung_to_ladder(ns["logic"], ns["mapping"])
        assert list(original_bundle.main_rows) == list(reproduced_bundle.main_rows)



"""CircuitPython code generation (feature-complete v1)."""

from __future__ import annotations

import math as _math
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

from pyrung.circuitpy.codegen._constants import _TYPE_DEFAULTS
from pyrung.circuitpy.codegen.context import CodegenContext
from pyrung.circuitpy.codegen.render import _render_code
from pyrung.circuitpy.codegen.render_runtime import _render_runtime
from pyrung.circuitpy.hardware import P1AM
from pyrung.circuitpy.modbus import ModbusClientConfig, ModbusServerConfig
from pyrung.circuitpy.p1am import RunStopConfig, board
from pyrung.circuitpy.validation import validate_circuitpy_program
from pyrung.click.tag_map import TagMap
from pyrung.core.memory_block import Block
from pyrung.core.program import Program
from pyrung.core.system_points import SYSTEM_TAGS_BY_NAME, system
from pyrung.core.tag import Tag

MappedTagScope = Literal["referenced_only", "all_mapped"]

_DEFAULT_RUNSTOP = RunStopConfig()


@dataclass(frozen=True)
class CircuitPyOutput:
    """Result of :func:`generate_circuitpy`.

    *code* is the ``code.py`` content (program-specific, stays as ``.py``).
    *runtime* is the ``pyrung_rt.py`` content (generic runtime library,
    intended to be compiled to ``.mpy`` via ``mpy-cross``).  An empty
    string means no runtime module is needed.
    """

    code: str
    runtime: str


def _needs_modbus_backing(tag: Tag, mode: MappedTagScope) -> bool:
    if mode == "all_mapped":
        return True
    return tag.default != _TYPE_DEFAULTS[tag.type]


def generate_circuitpy(
    program: Program,
    hw: P1AM,
    *,
    target_scan_ms: float,
    watchdog_ms: int | None = None,
    runstop: RunStopConfig | None = _DEFAULT_RUNSTOP,
    modbus_server: ModbusServerConfig | None = None,
    modbus_client: ModbusClientConfig | None = None,
    tag_map: TagMap | None = None,
    mapped_tag_scope: MappedTagScope = "referenced_only",
) -> CircuitPyOutput:
    if not isinstance(program, Program):
        raise TypeError(f"program must be Program, got {type(program).__name__}")
    if not isinstance(hw, P1AM):
        raise TypeError(f"hw must be P1AM, got {type(hw).__name__}")
    if not isinstance(target_scan_ms, (int, float)):
        raise TypeError(
            f"target_scan_ms must be a finite number > 0, got {type(target_scan_ms).__name__}"
        )
    if not _math.isfinite(float(target_scan_ms)) or float(target_scan_ms) <= 0:
        raise ValueError("target_scan_ms must be finite and > 0")
    if watchdog_ms is not None:
        if not isinstance(watchdog_ms, int):
            raise TypeError(f"watchdog_ms must be int or None, got {type(watchdog_ms).__name__}")
        if watchdog_ms < 0:
            raise ValueError("watchdog_ms must be >= 0")
    if runstop is not None and not isinstance(runstop, RunStopConfig):
        raise TypeError(f"runstop must be RunStopConfig or None, got {type(runstop).__name__}")
    if modbus_server is not None and not isinstance(modbus_server, ModbusServerConfig):
        raise TypeError(
            f"modbus_server must be ModbusServerConfig or None, got {type(modbus_server).__name__}"
        )
    if modbus_client is not None and not isinstance(modbus_client, ModbusClientConfig):
        raise TypeError(
            f"modbus_client must be ModbusClientConfig or None, got {type(modbus_client).__name__}"
        )
    if tag_map is not None and not isinstance(tag_map, TagMap):
        raise TypeError(f"tag_map must be TagMap or None, got {type(tag_map).__name__}")
    if mapped_tag_scope not in {"referenced_only", "all_mapped"}:
        raise ValueError("mapped_tag_scope must be 'referenced_only' or 'all_mapped'")
    if (modbus_server is not None or modbus_client is not None) and tag_map is None:
        raise ValueError("tag_map is required when modbus_server or modbus_client is enabled")

    ctx = CodegenContext(
        program=program,
        hw=hw,
        target_scan_ms=float(target_scan_ms),
        watchdog_ms=watchdog_ms,
        modbus_server=modbus_server,
        modbus_client=modbus_client,
        tag_map=tag_map,
    )
    ctx.collect_hw_bindings()
    ctx.collect_program_references()
    ctx.runstop = runstop
    if runstop is not None:
        ctx.ensure_tag_referenced(board.switch)
        if runstop.expose_mode_tags:
            ctx.ensure_tag_referenced(system.sys.mode_run)
            ctx.ensure_tag_referenced(system.sys.cmd_mode_stop)
    if tag_map is not None:
        for entry in tag_map.entries:
            logical = getattr(entry, "logical", None)
            if isinstance(logical, Tag):
                if _needs_modbus_backing(logical, mapped_tag_scope):
                    ctx.ensure_tag_referenced(logical)
                continue
            if isinstance(logical, Block):
                ctx._ensure_block_binding(logical)
                for addr in getattr(entry, "logical_addresses", ()):
                    tag = logical[addr]
                    if _needs_modbus_backing(tag, mapped_tag_scope):
                        ctx.ensure_tag_referenced(tag)
        if mapped_tag_scope == "all_mapped":
            for slot in tag_map.mapped_slots():
                if slot.source != "system":
                    continue
                system_tag = SYSTEM_TAGS_BY_NAME.get(slot.logical_name)
                if system_tag is not None:
                    ctx.ensure_tag_referenced(system_tag)

    if not hw._slots and not ctx.board_tag_names:
        raise ValueError(
            "P1AM hardware config must include at least one configured slot or referenced board tags"
        )

    if hw._slots:
        slot_numbers = sorted(hw._slots)
        expected = list(range(1, slot_numbers[-1] + 1))
        if slot_numbers != expected:
            raise ValueError(
                "Configured slots must be contiguous from 1..N for v1 roll-call generation"
            )

    report = validate_circuitpy_program(program, hw=hw, mode="strict")
    if report.errors:
        lines = [f"{len(report.errors)} error(s)."]
        for err in report.errors:
            lines.append(f"{err.code} @ {err.location}: {err.message}")
        raise ValueError("\n".join(lines))

    ctx.collect_retentive_tags()
    ctx.assign_symbols()

    # Predict has_runtime from config — modbus requires a runtime module.
    # Render code first so compile_rung populates ctx.used_helpers and
    # ctx.modbus_client_specs; _render_runtime needs those.
    has_runtime = ctx.modbus_server is not None or ctx.modbus_client is not None
    source = _render_code(ctx, has_runtime=has_runtime)
    runtime_source = _render_runtime(ctx) if has_runtime else ""

    if runtime_source:
        try:
            compile(runtime_source, "pyrung_rt.py", "exec")
        except SyntaxError as exc:
            raise RuntimeError(f"Generated runtime source is invalid: {exc}") from exc
    try:
        compile(source, "code.py", "exec")
    except SyntaxError as exc:
        raise RuntimeError(f"Generated source is invalid: {exc}") from exc
    return CircuitPyOutput(code=source, runtime=runtime_source)


def write_circuitpy(
    program: Program,
    hw: P1AM,
    *,
    output_dir: str | Path,
    target_scan_ms: float,
    watchdog_ms: int | None = None,
    runstop: RunStopConfig | None = _DEFAULT_RUNSTOP,
    modbus_server: ModbusServerConfig | None = None,
    modbus_client: ModbusClientConfig | None = None,
    tag_map: TagMap | None = None,
    mapped_tag_scope: MappedTagScope = "referenced_only",
) -> Path:
    """Generate and write ``code.py`` (and ``pyrung_rt.py`` when needed) to *output_dir*.

    Accepts the same parameters as :func:`generate_circuitpy` plus
    ``output_dir``.  Returns the path to the written ``code.py``.
    """
    result = generate_circuitpy(
        program,
        hw,
        target_scan_ms=target_scan_ms,
        watchdog_ms=watchdog_ms,
        runstop=runstop,
        modbus_server=modbus_server,
        modbus_client=modbus_client,
        tag_map=tag_map,
        mapped_tag_scope=mapped_tag_scope,
    )
    out = Path(output_dir)
    code_path = out / "code.py"
    code_path.write_text(result.code, encoding="utf-8")
    if result.runtime:
        runtime_path = out / "pyrung_rt.py"
        runtime_path.write_text(result.runtime, encoding="utf-8")
    return code_path



"""Stack/evaluate ownership for DAP command handling.

Owns stack frame/scopes/variables/evaluate payload construction.
Must preserve evaluate command semantics and error message text.
"""

from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from pyrung.core.state import SystemState
from pyrung.dap.expressions import And, Compare, Expr, ExpressionParseError, Not, Or
from pyrung.dap.expressions import compile as compile_condition
from pyrung.dap.expressions import parse as parse_condition

HandlerResult = tuple[dict[str, Any], list[tuple[str, dict[str, Any] | None]]]

_SIMPLE_ATTR_RE = re.compile(r"^([A-Za-z_][A-Za-z0-9_]*)\.([A-Za-z_][A-Za-z0-9_]*)$")
_INDEXED_TAG_RE = re.compile(r"^([A-Za-z_][A-Za-z0-9_]*)\[(\d+)\]$")
_INSTANCE_FIELD_RE = re.compile(r"^([A-Za-z_][A-Za-z0-9_]*)\[(\d+)\]\.([A-Za-z_][A-Za-z0-9_]*)$")
_FIELD_INDEX_RE = re.compile(r"^([A-Za-z_][A-Za-z0-9_]*)\.([A-Za-z_][A-Za-z0-9_]*)\[(\d+)\]$")


@dataclass(frozen=True)
class _VariablesRequestArgs:
    variablesReference: Any = 0


@dataclass(frozen=True)
class _EvaluateRequestArgs:
    expression: Any = None
    context: Any = None


def on_stack_trace(adapter: Any, args: dict[str, Any]) -> HandlerResult:
    with adapter._state_lock:
        runner = adapter._require_runner_locked()
        rungs = adapter._top_level_rungs(runner)
        current_step = adapter._current_step
        current_index = adapter._current_rung_index
        program_path = adapter._program_path

    start_frame = int(args.get("startFrame", 0))
    levels = int(args.get("levels", 0))

    if current_step is not None:
        frames = adapter._formatter.build_current_stack_frames(
            current_step=current_step,
            rungs=rungs,
            subroutine_source_map=adapter._breakpoints.subroutine_sources(),
            canonical_path=adapter._canonical_path,
        )
    elif rungs:
        order = list(range(len(rungs)))
        if current_index is not None and 0 <= current_index < len(rungs):
            order = [current_index, *[i for i in order if i != current_index]]

        frames = [
            adapter._formatter.stack_frame_from_rung(
                frame_id=idx,
                name=f"Rung {idx}",
                rung=rungs[idx],
            )
            for idx in order
        ]
    else:
        frame: dict[str, Any] = {"id": 0, "name": "Scan", "line": 1, "column": 1}
        if program_path is not None:
            frame["source"] = {"name": Path(program_path).name, "path": program_path}
        frames = [frame]

    if levels > 0:
        visible = frames[start_frame : start_frame + levels]
    else:
        visible = frames[start_frame:]

    return {"stackFrames": visible, "totalFrames": len(frames)}, []


def on_scopes(adapter: Any, _args: dict[str, Any]) -> HandlerResult:
    scopes = [
        {"name": "Tags", "variablesReference": adapter.TAGS_SCOPE_REF, "expensive": False},
        {"name": "Forces", "variablesReference": adapter.FORCES_SCOPE_REF, "expensive": False},
        {"name": "Memory", "variablesReference": adapter.MEMORY_SCOPE_REF, "expensive": False},
    ]
    with adapter._state_lock:
        has_monitors = bool(adapter._monitor_handles)
        monitor_count = len(adapter._monitor_handles)
    if has_monitors:
        scopes.append(
            {
                "name": "PLC Monitors",
                "variablesReference": adapter._monitor_scope_ref,
                "expensive": False,
                "namedVariables": monitor_count,
            }
        )
    return {"scopes": scopes}, []


def on_variables(adapter: Any, args: dict[str, Any]) -> HandlerResult:
    parsed = adapter._parse_request_args(_VariablesRequestArgs, args)
    ref = int(parsed.variablesReference)
    with adapter._state_lock:
        runner = adapter._require_runner_locked()
        current_ctx = adapter._current_ctx
        state = runner.current_state
        forces = dict(runner.forces)

    if ref == adapter.TAGS_SCOPE_REF:
        values = dict(state.tags)
        if current_ctx is not None:
            values.update(getattr(current_ctx, "_tags_pending", {}))
        return {"variables": adapter._as_dap_variables(values)}, []

    if ref == adapter.FORCES_SCOPE_REF:
        return {"variables": adapter._as_dap_variables(forces)}, []

    if ref == adapter.MEMORY_SCOPE_REF:
        values = dict(state.memory)
        if current_ctx is not None:
            values.update(getattr(current_ctx, "_memory_pending", {}))
        return {"variables": adapter._as_dap_variables(values)}, []

    if ref == adapter._monitor_scope_ref:
        variables = adapter._monitor_variables()
        return {"variables": variables}, []

    raise adapter.DAPAdapterError(f"Unknown variablesReference: {ref}")


def on_evaluate(adapter: Any, args: dict[str, Any]) -> HandlerResult:
    parsed = adapter._parse_request_args(_EvaluateRequestArgs, args)
    expression = parsed.expression
    if not isinstance(expression, str) or not expression.strip():
        raise adapter.DAPAdapterError("evaluate.expression is required")
    context = parsed.context
    evaluate_context = context if isinstance(context, str) else "repl"

    with adapter._state_lock:
        adapter._require_runner_locked()
        if evaluate_context == "watch":
            result = adapter._evaluate_watch_expression_locked(expression)
        else:
            result = adapter._evaluate_repl_command_locked(expression)

    return {"result": adapter._format_value(result), "variablesReference": 0}, []


def evaluate_repl_command_locked(adapter: Any, expression: str) -> str:
    parts = expression.strip().split()
    command = parts[0]
    runner = adapter._require_runner_locked()
    if command == "force":
        if len(parts) < 3:
            raise adapter.DAPAdapterError("Usage: force <tag> <value>")
        tag = parts[1]
        raw_value = expression.strip().split(None, 2)[2]
        value = adapter._parse_literal(raw_value)
        runner.add_force(tag, value)
        return f"Forced {tag}={value!r}"
    if command in {"remove_force", "unforce"}:
        if len(parts) != 2:
            raise adapter.DAPAdapterError("Usage: remove_force <tag>")
        tag = parts[1]
        runner.remove_force(tag)
        return f"Removed force {tag}"
    if command == "clear_forces":
        runner.clear_forces()
        return "Cleared all forces"
    raise adapter.DAPAdapterError(
        "Unsupported Debug Console command. Use force/remove_force/unforce/clear_forces. "
        "Use Watch for predicate expressions."
    )


def evaluate_watch_expression_locked(adapter: Any, expression: str) -> Any:
    state = adapter._effective_evaluate_state_locked()
    try:
        parsed = parse_condition(expression)
    except ExpressionParseError as exc:
        raise adapter.DAPAdapterError(str(exc)) from exc

    references = adapter._expression_references(parsed)
    for name in sorted(references):
        if not adapter._state_has_reference(state, name):
            raise adapter.DAPAdapterError(f"Unknown tag or memory reference: {name}")

    if isinstance(parsed, Compare) and parsed.op is None and parsed.right is None:
        return adapter._state_value_for_reference(state, parsed.tag.name)

    state_for_eval = _state_with_reference_aliases(adapter, state, references)
    return compile_condition(parsed)(state_for_eval)


def effective_evaluate_state_locked(adapter: Any) -> SystemState:
    runner = adapter._require_runner_locked()
    state = runner.current_state
    current_ctx = adapter._current_ctx
    if current_ctx is None:
        return state

    pending_tags = getattr(current_ctx, "_tags_pending", {})
    if isinstance(pending_tags, dict) and pending_tags:
        state = state.with_tags(dict(pending_tags))

    pending_memory = getattr(current_ctx, "_memory_pending", {})
    if isinstance(pending_memory, dict) and pending_memory:
        state = state.with_memory(dict(pending_memory))

    return state


def expression_references(adapter: Any, expr: Expr) -> set[str]:
    if isinstance(expr, Compare):
        return {expr.tag.name}
    if isinstance(expr, Not):
        return {expr.child.name}
    if isinstance(expr, And):
        refs: set[str] = set()
        for child in expr.children:
            refs.update(adapter._expression_references(child))
        return refs
    if isinstance(expr, Or):
        refs: set[str] = set()
        for child in expr.children:
            refs.update(adapter._expression_references(child))
        return refs
    return set()


def state_has_reference(adapter: Any, state: SystemState, name: str) -> bool:
    return _resolve_reference_name(adapter, state, name) is not None


def state_value_for_reference(adapter: Any, state: SystemState, name: str) -> Any:
    resolved = _resolve_reference_name(adapter, state, name)
    if resolved is None:
        return None
    if resolved in state.tags:
        return state.tags.get(resolved)
    if resolved in state.memory:
        return state.memory.get(resolved)
    return None


def _state_with_reference_aliases(
    adapter: Any,
    state: SystemState,
    references: set[str],
) -> SystemState:
    tag_aliases: dict[str, Any] = {}
    memory_aliases: dict[str, Any] = {}
    for reference in references:
        if reference in state.tags or reference in state.memory:
            continue
        resolved = _resolve_reference_name(adapter, state, reference)
        if resolved is None:
            continue
        if resolved in state.tags:
            tag_aliases[reference] = state.tags.get(resolved)
            continue
        if resolved in state.memory:
            memory_aliases[reference] = state.memory.get(resolved)

    if tag_aliases:
        state = state.with_tags(tag_aliases)
    if memory_aliases:
        state = state.with_memory(memory_aliases)
    return state


def _resolve_reference_name(adapter: Any, state: SystemState, name: str) -> str | None:
    reference = name.strip()
    if not reference:
        return None

    if reference in state.tags or reference in state.memory:
        return reference

    for candidate in _reference_alias_candidates(adapter, state, reference):
        if candidate in state.tags or candidate in state.memory:
            return candidate
    return None


def _reference_alias_candidates(adapter: Any, state: SystemState, reference: str) -> list[str]:
    del adapter

    candidates: list[str] = []

    simple_attr = _SIMPLE_ATTR_RE.match(reference)
    if simple_attr is not None:
        root_name = simple_attr.group(1)
        leaf_name = simple_attr.group(2)
        if root_name and root_name[0].isupper():
            _append_candidate(candidates, f"{root_name}_{leaf_name}")
            _append_candidate(candidates, leaf_name)

    indexed_tag = _INDEXED_TAG_RE.match(reference)
    if indexed_tag is not None:
        _append_candidate(candidates, f"{indexed_tag.group(1)}{indexed_tag.group(2)}")

    instance_field = _INSTANCE_FIELD_RE.match(reference)
    if instance_field is not None:
        _append_struct_candidates(
            candidates,
            state,
            base=instance_field.group(1),
            index=instance_field.group(2),
            field=instance_field.group(3),
        )

    field_index = _FIELD_INDEX_RE.match(reference)
    if field_index is not None:
        _append_struct_candidates(
            candidates,
            state,
            base=field_index.group(1),
            index=field_index.group(3),
            field=field_index.group(2),
        )

    return candidates


def _append_struct_candidates(
    candidates: list[str],
    state: SystemState,
    *,
    base: str,
    index: str,
    field: str,
) -> None:
    _append_candidate(candidates, f"{base}{index}_{field}")

    suffix = f"{index}_{field}"
    suffix_matches = [tag_name for tag_name in state.tags if tag_name.endswith(suffix)]
    if len(suffix_matches) == 1:
        _append_candidate(candidates, suffix_matches[0])


def _append_candidate(candidates: list[str], value: str) -> None:
    if value and value not in candidates:
        candidates.append(value)



"""laddercodec CSV → pyrung source code generator."""

from __future__ import annotations

from pyrung.click.codegen.api import ladder_to_pyrung, ladder_to_pyrung_project

__all__ = ["ladder_to_pyrung", "ladder_to_pyrung_project"]



"""PLCRunner - Generator-driven PLC execution engine.

The runner orchestrates scan cycle execution with inversion of control.
The consumer drives execution via step(), allowing input injection,
inspection, and pause at any point.

Uses ScanContext to batch all tag/memory updates within a scan cycle,
reducing object allocation from O(instructions) to O(1) per scan.
"""

from __future__ import annotations

import time
from collections.abc import Callable, Iterable, Mapping
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Literal

from pyrung.core.condition_trace import ConditionTraceEngine
from pyrung.core.context import ConditionView, ScanContext
from pyrung.core.debug_trace import RungTrace, RungTraceEvent, TraceEvent
from pyrung.core.debugger import PLCDebugger
from pyrung.core.history import History
from pyrung.core.input_overrides import InputOverrideManager
from pyrung.core.live_binding import reset_active_runner, set_active_runner
from pyrung.core.state import SystemState
from pyrung.core.system_points import (
    _BATTERY_PRESENT_KEY,
    _MODE_RUN_KEY,
    SYSTEM_TAGS_BY_NAME,
    SystemPointRuntime,
)
from pyrung.core.time_mode import TimeMode
from pyrung.core.trace_formatter import TraceFormatter

if TYPE_CHECKING:
    from collections.abc import Generator, Iterator

    from pyrung.core.condition import Condition
    from pyrung.core.rung import Rung
    from pyrung.core.tag import Tag


@dataclass(frozen=True)
class ScanStep:
    """Debug scan step emitted at rung boundaries."""

    rung_index: int
    rung: Rung
    ctx: ScanContext
    kind: Literal["rung", "branch", "subroutine", "instruction"]
    subroutine_name: str | None
    depth: int
    call_stack: tuple[str, ...]
    source_file: str | None
    source_line: int | None
    end_line: int | None
    enabled_state: Literal["enabled", "disabled_local", "disabled_parent"] | None
    trace: TraceEvent | None
    instruction_kind: str | None


@dataclass
class _MonitorRegistration:
    id: int
    tag_name: str
    callback: Callable[[Any, Any], None]
    enabled: bool = True
    removed: bool = False


@dataclass
class _BreakpointRegistration:
    id: int
    predicate: Callable[[SystemState], bool]
    action: Literal["pause", "snapshot"]
    label: str | None = None
    enabled: bool = True
    removed: bool = False


class _RunnerHandle:
    """Mutable registration handle used by monitor/breakpoint APIs."""

    __slots__ = ("_id", "_remove", "_enable", "_disable")

    def __init__(
        self,
        *,
        handle_id: int,
        remove: Callable[[int], None],
        enable: Callable[[int], None],
        disable: Callable[[int], None],
    ) -> None:
        self._id = handle_id
        self._remove = remove
        self._enable = enable
        self._disable = disable

    @property
    def id(self) -> int:
        return self._id

    def remove(self) -> None:
        self._remove(self._id)

    def enable(self) -> None:
        self._enable(self._id)

    def disable(self) -> None:
        self._disable(self._id)


class _BreakpointBuilder:
    """Fluent builder returned by ``runner.when(...)`` / ``runner.when_fn(...)``."""

    __slots__ = ("_runner", "_predicate")

    def __init__(self, runner: PLCRunner, predicate: Callable[[SystemState], bool]) -> None:
        self._runner = runner
        self._predicate = predicate

    def pause(self) -> _RunnerHandle:
        return self._runner._register_breakpoint(
            predicate=self._predicate,
            action="pause",
            label=None,
        )

    def snapshot(self, label: str) -> _RunnerHandle:
        return self._runner._register_breakpoint(
            predicate=self._predicate,
            action="snapshot",
            label=label,
        )


def _iter_referenced_tags(root: Any) -> tuple[Tag, ...]:
    """Collect Tag objects reachable from a logic object graph."""
    from pyrung.core.tag import Tag as TagClass

    found_by_name: dict[str, TagClass] = {}
    visited: set[int] = set()
    queue: list[Any] = [root]

    while queue:
        current = queue.pop()
        if current is None:
            continue
        if isinstance(current, TagClass):
            found_by_name[current.name] = current
            continue
        if isinstance(current, (str, bytes, bytearray, int, float, bool)):
            continue

        current_id = id(current)
        if current_id in visited:
            continue
        visited.add(current_id)

        if isinstance(current, Mapping):
            queue.extend(current.keys())
            queue.extend(current.values())
            continue
        if isinstance(current, tuple | list | set | frozenset):
            queue.extend(current)
            continue

        if hasattr(current, "__dict__"):
            queue.extend(vars(current).values())
            continue
        if hasattr(current, "__slots__"):
            for slot in current.__slots__:
                if slot in {"__weakref__", "__dict__"}:
                    continue
                if hasattr(current, slot):
                    queue.append(getattr(current, slot))

    return tuple(found_by_name.values())


class PLCRunner:
    """Generator-driven PLC execution engine.

    Executes PLC logic as pure functions: Logic(state) -> new_state.
    The consumer controls execution via step(), enabling:
    - Input injection via patch()
    - Inspection of retained historical state via runner.history
    - Pause/resume at any scan boundary

    Attributes:
        current_state: The current SystemState snapshot.
        history: Query interface for retained SystemState snapshots.
        simulation_time: Current simulation clock (seconds).
        time_mode: Current time mode (FIXED_STEP or REALTIME).
    """

    def __init__(
        self,
        logic: list[Any] | Any = None,
        initial_state: SystemState | None = None,
        *,
        history_limit: int | None = None,
    ) -> None:
        """Create a new PLCRunner.

        Args:
            logic: Program, list of rungs, or None for empty logic.
            initial_state: Starting state. Defaults to SystemState().
            history_limit: Max retained snapshots including initial state.
                Use None for unbounded history.
        """
        if history_limit is not None and history_limit < 1:
            raise ValueError("history_limit must be >= 1 or None")

        self._logic: list[Rung]
        # Handle different logic types
        # Import Program here to avoid circular import at module level
        from pyrung.core.program import Program

        if logic is None:
            self._logic = []
        elif isinstance(logic, Program):
            self._logic = logic.rungs
        elif isinstance(logic, list):
            self._logic = logic
        else:
            self._logic = [logic]

        self._state = initial_state if initial_state is not None else SystemState()
        self._history_limit = history_limit
        self._running = True
        self._battery_present = True
        self._state = self._apply_runtime_memory_flags(
            self._state,
            mode_run=self._running,
            battery_present=self._battery_present,
        )
        self._history = History(self._state, limit=history_limit)
        self._rung_traces_by_scan: dict[int, dict[int, RungTrace]] = {}
        self._inflight_scan_id: int | None = None
        self._inflight_rung_events: dict[int, list[RungTraceEvent]] = {}
        self._latest_inflight_trace_event: tuple[int, int, RungTraceEvent] | None = None
        self._latest_committed_trace_event: tuple[int, int, RungTraceEvent] | None = None
        self._playhead = self._state.scan_id
        self._time_mode = TimeMode.FIXED_STEP
        self._dt = 0.1  # Default: 100ms per scan
        self._last_step_time: float | None = None  # For REALTIME mode
        self._rtc_base = self._normalize_rtc_datetime(datetime.now())
        self._rtc_base_sim_time = float(self._state.timestamp)
        self._system_runtime = SystemPointRuntime(
            time_mode_getter=lambda: self._time_mode,
            fixed_step_dt_getter=lambda: self._dt,
            rtc_now_getter=self._rtc_at_sim_time,
            rtc_setter=self._set_rtc_internal,
        )
        self._input_overrides = InputOverrideManager(is_read_only=self._system_runtime.is_read_only)
        # Preserve direct access used in tests/live-tag helpers.
        self._pending_patches = self._input_overrides.pending_patches
        self._forces = self._input_overrides.forces_mutable
        self._condition_trace = ConditionTraceEngine(formatter=TraceFormatter())
        self._debugger = PLCDebugger(step_factory=ScanStep)
        self._next_debug_handle_id = 1
        self._monitors_by_id: dict[int, _MonitorRegistration] = {}
        self._breakpoints_by_id: dict[int, _BreakpointRegistration] = {}
        self._pause_requested_this_scan = False
        self._known_tags_by_name: dict[str, Tag] = {}
        self._refresh_known_tags_from_logic()

    @property
    def current_state(self) -> SystemState:
        """Current state snapshot."""
        return self._state

    @property
    def history(self) -> History:
        """Read-only history query surface."""
        return self._history

    @property
    def playhead(self) -> int:
        """Current scan id used for inspection/time-travel queries."""
        return self._playhead

    def seek(self, scan_id: int) -> SystemState:
        """Move playhead to a retained scan and return that snapshot."""
        state = self.history.at(scan_id)
        self._playhead = state.scan_id
        return state

    def rewind(self, seconds: float) -> SystemState:
        """Move playhead backward in time by ``seconds`` and return snapshot."""
        if seconds < 0:
            raise ValueError("seconds must be >= 0")

        playhead_state = self.history.at(self._playhead)
        target_timestamp = playhead_state.timestamp - seconds

        target_state = self.history.at_or_before_timestamp(target_timestamp)
        if target_state is None:
            target_state = self.history.at(self.history.oldest_scan_id)

        self._playhead = target_state.scan_id
        return target_state

    def diff(self, scan_a: int, scan_b: int) -> dict[str, tuple[Any, Any]]:
        """Return changed tag values between two retained historical scans."""
        state_a = self.history.at(scan_a)
        state_b = self.history.at(scan_b)

        changed: dict[str, tuple[Any, Any]] = {}
        all_keys = sorted(set(state_a.tags.keys()) | set(state_b.tags.keys()))
        for key in all_keys:
            old_value = state_a.tags.get(key)
            new_value = state_b.tags.get(key)
            if old_value != new_value:
                changed[key] = (old_value, new_value)
        return changed

    def inspect(self, rung_id: int, scan_id: int | None = None) -> RungTrace:
        """Return retained rung-level debug trace for one scan.

        If ``scan_id`` is omitted, the current playhead scan is inspected.
        Raises:
            KeyError: Missing scan id, or missing rung trace for retained scan.
        """
        target_scan_id = self._playhead if scan_id is None else scan_id
        self.history.at(target_scan_id)

        scan_traces = self._rung_traces_by_scan.get(target_scan_id)
        if scan_traces is None:
            raise KeyError(rung_id)

        try:
            return scan_traces[rung_id]
        except KeyError as exc:
            raise KeyError(rung_id) from exc

    def inspect_event(self) -> tuple[int, int, RungTraceEvent] | None:
        """Return the latest debug-trace event for active/committed debug-path scans.

        Returns:
            A tuple of ``(scan_id, rung_id, event)``. In-flight debug-scan events
            are preferred when available. Otherwise, the latest retained committed
            debug-scan event is returned.

        Notes:
            - This API is populated by ``scan_steps_debug()`` only.
            - Scans produced through ``step()/run()/run_for()/run_until()`` do not
              contribute trace events here.
        """
        inflight = self._latest_inflight_trace_event
        if self._inflight_scan_id is not None and inflight is not None:
            return inflight

        committed = self._latest_committed_trace_event
        if committed is None:
            return None

        scan_id, rung_id, _event = committed
        trace = self._rung_traces_by_scan.get(scan_id, {}).get(rung_id)
        if trace is None:
            self._latest_committed_trace_event = None
            return None

        if not trace.events:
            self._latest_committed_trace_event = None
            return None

        latest_event = trace.events[-1]
        self._latest_committed_trace_event = (scan_id, rung_id, latest_event)
        return self._latest_committed_trace_event

    def fork(self, scan_id: int | None = None) -> PLCRunner:
        """Create an independent runner from retained historical state.

        Args:
            scan_id: Snapshot to fork from. Defaults to current committed tip state.
        """
        target_scan_id = self._state.scan_id if scan_id is None else scan_id
        historical_state = self.history.at(target_scan_id)
        fork = PLCRunner(
            logic=list(self._logic),
            initial_state=historical_state,
            history_limit=self._history_limit,
        )
        fork.set_time_mode(self._time_mode, dt=self._dt)
        parent_rtc_at_fork_point = self.system_runtime._rtc_now(historical_state)
        fork._set_rtc_internal(parent_rtc_at_fork_point, fork.current_state.timestamp)
        return fork

    def fork_from(self, scan_id: int) -> PLCRunner:
        """Create an independent runner from a retained historical snapshot."""
        return self.fork(scan_id=scan_id)

    @property
    def simulation_time(self) -> float:
        """Current simulation clock in seconds."""
        return self._state.timestamp

    @property
    def time_mode(self) -> TimeMode:
        """Current time mode."""
        return self._time_mode

    @property
    def system_runtime(self) -> SystemPointRuntime:
        """System point runtime component."""
        return self._system_runtime

    def stop(self) -> None:
        """Transition PLC to STOP mode."""
        if not self._running:
            return
        self._running = False
        self._state = self._apply_runtime_memory_flags(
            self._state,
            mode_run=False,
            battery_present=self._battery_present,
        )

    def set_battery_present(self, value: bool) -> None:
        """Configure simulated backup battery presence."""
        self._battery_present = bool(value)
        self._state = self._apply_runtime_memory_flags(
            self._state,
            mode_run=self._running,
            battery_present=self._battery_present,
        )

    def reboot(self) -> SystemState:
        """Power-cycle the runner and return the reset state."""
        tag_values = self._rebuild_tags_for_reset(
            preserve_all=self._battery_present,
            preserve_retentive=False,
        )
        self._reset_runtime_scope(
            tag_values=tag_values,
            mode_run=True,
            preserve_rtc_continuity=self._battery_present,
        )
        self._running = True
        return self._state

    def set_rtc(self, value: datetime) -> None:
        """Set the current RTC value for the runner."""
        self._set_rtc_internal(self._normalize_rtc_datetime(value), self._state.timestamp)

    def set_time_mode(self, mode: TimeMode, *, dt: float = 0.1) -> None:
        """Set the time mode for simulation.

        Args:
            mode: TimeMode.FIXED_STEP or TimeMode.REALTIME.
            dt: Time delta per scan (only used for FIXED_STEP mode).
        """
        self._time_mode = mode
        self._dt = dt
        if mode == TimeMode.REALTIME:
            self._last_step_time = time.perf_counter()

    @staticmethod
    def _apply_runtime_memory_flags(
        state: SystemState,
        *,
        mode_run: bool,
        battery_present: bool,
    ) -> SystemState:
        memory = state.memory.set(_MODE_RUN_KEY, bool(mode_run)).set(
            _BATTERY_PRESENT_KEY, bool(battery_present)
        )
        return state.set(memory=memory)

    def _refresh_known_tags_from_logic(self) -> None:
        for rung in self._logic:
            for tag in _iter_referenced_tags(rung):
                self._register_known_tag(tag)

    def _register_known_tag(self, tag: Tag) -> None:
        if tag.name in SYSTEM_TAGS_BY_NAME:
            return
        existing = self._known_tags_by_name.get(tag.name)
        if existing is None:
            self._known_tags_by_name[tag.name] = tag
            return
        if (
            existing.type != tag.type
            or existing.retentive != tag.retentive
            or existing.default != tag.default
        ):
            raise ValueError(
                f"Conflicting tag metadata for {tag.name!r}: existing "
                f"(type={existing.type.name}, retentive={existing.retentive}, "
                f"default={existing.default!r}) vs new "
                f"(type={tag.type.name}, retentive={tag.retentive}, default={tag.default!r})."
            )

    def _register_known_tags_from_mapping_keys(
        self,
        tags: Mapping[str, bool | int | float | str]
        | Mapping[Tag, bool | int | float | str]
        | Mapping[str | Tag, bool | int | float | str],
    ) -> None:
        from pyrung.core.tag import Tag as TagClass

        for key in tags:
            if isinstance(key, TagClass):
                self._register_known_tag(key)

    def _rebuild_tags_for_reset(
        self,
        *,
        preserve_all: bool,
        preserve_retentive: bool,
    ) -> dict[str, Any]:
        current_tags = self._state.tags
        rebuilt: dict[str, Any] = {}
        for tag in self._known_tags_by_name.values():
            if preserve_all:
                rebuilt[tag.name] = current_tags.get(tag.name, tag.default)
                continue
            if preserve_retentive and tag.retentive and tag.name in current_tags:
                rebuilt[tag.name] = current_tags[tag.name]
                continue
            rebuilt[tag.name] = tag.default
        return rebuilt

    def _clear_retained_debug_trace_caches(self) -> None:
        self._rung_traces_by_scan.clear()
        self._clear_inflight_debug_scan()
        self._latest_committed_trace_event = None

    def _normalize_rtc_datetime(self, value: datetime) -> datetime:
        if value.tzinfo is None:
            return value
        return value.astimezone().replace(tzinfo=None)

    def _set_rtc_internal(self, value: datetime, sim_time: float) -> None:
        self._rtc_base = value
        self._rtc_base_sim_time = float(sim_time)

    def _rtc_at_sim_time(self, sim_time: float) -> datetime:
        return self._rtc_base + timedelta(seconds=float(sim_time) - self._rtc_base_sim_time)

    def _reset_runtime_scope(
        self,
        *,
        tag_values: Mapping[str, Any],
        mode_run: bool,
        preserve_rtc_continuity: bool = True,
    ) -> None:
        rtc_after_reset = (
            self._rtc_at_sim_time(self._state.timestamp)
            if preserve_rtc_continuity
            else self._normalize_rtc_datetime(datetime.now())
        )
        next_state = SystemState().with_tags(dict(tag_values))
        self._state = self._apply_runtime_memory_flags(
            next_state,
            mode_run=mode_run,
            battery_present=self._battery_present,
        )
        self._set_rtc_internal(rtc_after_reset, self._state.timestamp)
        self._history = History(self._state, limit=self._history_limit)
        self._playhead = self._state.scan_id

        self._pending_patches.clear()
        self._forces.clear()
        self._pause_requested_this_scan = False
        self._clear_retained_debug_trace_caches()

        if self._time_mode == TimeMode.REALTIME:
            self._last_step_time = time.perf_counter()
        else:
            self._last_step_time = None

    def _stop_to_run_transition(self) -> None:
        if self._running:
            return
        tag_values = self._rebuild_tags_for_reset(
            preserve_all=False,
            preserve_retentive=True,
        )
        self._reset_runtime_scope(
            tag_values=tag_values,
            mode_run=True,
            preserve_rtc_continuity=True,
        )
        self._running = True

    def _ensure_running(self) -> None:
        if not self._running:
            self._stop_to_run_transition()

    def _sync_runtime_flags_from_state(self) -> None:
        self._running = bool(self._state.memory.get(_MODE_RUN_KEY, True))
        self._battery_present = bool(
            self._state.memory.get(_BATTERY_PRESENT_KEY, self._battery_present)
        )

    @contextmanager
    def active(self) -> Iterator[PLCRunner]:
        """Bind this runner as active for live Tag.value access."""
        token = set_active_runner(self)
        try:
            yield self
        finally:
            reset_active_runner(token)

    def patch(
        self,
        tags: Mapping[str, bool | int | float | str]
        | Mapping[Tag, bool | int | float | str]
        | Mapping[str | Tag, bool | int | float | str],
    ) -> None:
        """Queue tag values for next scan (one-shot).

        Values are applied at the start of the next step() call,
        then cleared. Use for momentary inputs like button presses.

        Args:
            tags: Dict of tag names or Tag objects to values.
        """
        self._register_known_tags_from_mapping_keys(tags)
        self._input_overrides.patch(tags)

    def add_force(self, tag: str | Tag, value: bool | int | float | str) -> None:
        """Persistently override a tag value until explicitly removed.

        The forced value is applied at the pre-logic force pass (phase 3) and
        re-applied at the post-logic force pass (phase 5) every scan.  Logic
        may temporarily diverge the value mid-scan, but the post-logic pass
        restores it before outputs are written.

        Forces persist across scans until `remove_force()` or `clear_forces()`
        is called.  Multiple forces may be active simultaneously.

        If a tag is both patched and forced in the same scan, the force
        overwrites the patch during the pre-logic pass.

        Args:
            tag: Tag name or `Tag` object to override.
            value: Value to hold. Must be compatible with the tag's type.

        Raises:
            ValueError: If the tag is a read-only system point.
        """
        from pyrung.core.tag import Tag as TagClass

        if isinstance(tag, TagClass):
            self._register_known_tag(tag)
        self._input_overrides.add_force(tag, value)

    def remove_force(self, tag: str | Tag) -> None:
        """Remove a single persistent force override.

        After removal the tag resumes its logic-computed value starting
        from the next scan.

        Args:
            tag: Tag name or `Tag` object whose force to remove.
        """
        self._input_overrides.remove_force(tag)

    def clear_forces(self) -> None:
        """Remove all active persistent force overrides.

        All forced tags resume their logic-computed values starting
        from the next scan.
        """
        self._input_overrides.clear_forces()

    @contextmanager
    def force(
        self,
        overrides: Mapping[str, bool | int | float | str]
        | Mapping[Tag, bool | int | float | str]
        | Mapping[str | Tag, bool | int | float | str],
    ) -> Iterator[PLCRunner]:
        """Temporarily apply forces for the duration of the context.

        On entry, saves the current force map and adds the given overrides.
        On exit (normally or on exception), the exact previous force map is
        restored — forces that existed before the context are reinstated, and
        forces added inside the context are removed.

        Safe for nesting: inner ``force()`` contexts layer on top of outer ones
        without disrupting them.

        Args:
            overrides: Mapping of tag name / ``Tag`` object to forced value.

        Example:
            ```python
            with runner.force({"AutoMode": True, "Fault": False}):
                runner.run(5)
            # AutoMode and Fault forces released here
            ```
        """
        self._register_known_tags_from_mapping_keys(overrides)
        with self._input_overrides.force(overrides):
            yield self

    @property
    def forces(self) -> Mapping[str, bool | int | float | str]:
        """Read-only view of active persistent overrides."""
        return self._input_overrides.forces

    def monitor(self, tag: str | Tag, callback: Callable[[Any, Any], None]) -> _RunnerHandle:
        """Call ``callback(current, previous)`` after commit when tag value changes."""
        tag_name = self._normalize_tag_name(tag, method="monitor")
        monitor_id = self._next_handle_id()
        self._monitors_by_id[monitor_id] = _MonitorRegistration(
            id=monitor_id,
            tag_name=tag_name,
            callback=callback,
        )
        return _RunnerHandle(
            handle_id=monitor_id,
            remove=self._remove_monitor,
            enable=self._enable_monitor,
            disable=self._disable_monitor,
        )

    def when(
        self,
        *conditions: Condition | Tag | tuple[Condition | Tag, ...] | list[Condition | Tag],
    ) -> _BreakpointBuilder:
        """Create a condition breakpoint builder evaluated after each committed scan."""
        predicate = self._compile_condition_predicate(*conditions, method="when")
        return _BreakpointBuilder(self, predicate)

    def when_fn(self, predicate: Callable[[SystemState], bool]) -> _BreakpointBuilder:
        """Create a callable-predicate breakpoint builder evaluated after each scan."""
        return _BreakpointBuilder(self, predicate)

    def _compile_condition_predicate(
        self,
        *conditions: Condition | Tag | tuple[Condition | Tag, ...] | list[Condition | Tag],
        method: Literal["run_until", "when"],
    ) -> Callable[[SystemState], bool]:
        """Compile a Tag/Condition expression into a ``SystemState`` predicate."""
        if not conditions:
            raise TypeError(f"{method}() requires at least one condition")
        if len(conditions) == 1 and callable(conditions[0]):
            raise TypeError(
                f"{method}() now accepts Tag/Condition expressions; "
                f"use {method}_fn(...) for callable predicates."
            )

        from pyrung.core.condition import _as_condition, _normalize_and_condition

        normalized = _normalize_and_condition(
            *conditions,
            coerce=_as_condition,
            empty_error=f"{method}() requires at least one condition",
            group_empty_error=f"{method}() condition group cannot be empty",
        )

        def _predicate(state: SystemState) -> bool:
            ctx = ScanContext(
                state,
                resolver=self._system_runtime.resolve,
                read_only_tags=self._system_runtime.read_only_tags,
            )
            return normalized.evaluate(ctx)

        return _predicate

    def _normalize_tag_name(self, tag: str | Tag, *, method: str) -> str:
        from pyrung.core.tag import Tag as TagClass

        if isinstance(tag, TagClass):
            return tag.name
        if isinstance(tag, str):
            return tag
        raise TypeError(f"{method}() tag must be str or Tag, got {type(tag).__name__}")

    def _next_handle_id(self) -> int:
        handle_id = self._next_debug_handle_id
        self._next_debug_handle_id += 1
        return handle_id

    def _remove_monitor(self, monitor_id: int) -> None:
        registration = self._monitors_by_id.get(monitor_id)
        if registration is None or registration.removed:
            return
        registration.removed = True
        registration.enabled = False

    def _enable_monitor(self, monitor_id: int) -> None:
        registration = self._monitors_by_id.get(monitor_id)
        if registration is None or registration.removed:
            return
        registration.enabled = True

    def _disable_monitor(self, monitor_id: int) -> None:
        registration = self._monitors_by_id.get(monitor_id)
        if registration is None or registration.removed:
            return
        registration.enabled = False

    def _register_breakpoint(
        self,
        *,
        predicate: Callable[[SystemState], bool],
        action: Literal["pause", "snapshot"],
        label: str | None,
    ) -> _RunnerHandle:
        breakpoint_id = self._next_handle_id()
        self._breakpoints_by_id[breakpoint_id] = _BreakpointRegistration(
            id=breakpoint_id,
            predicate=predicate,
            action=action,
            label=label,
        )
        return _RunnerHandle(
            handle_id=breakpoint_id,
            remove=self._remove_breakpoint,
            enable=self._enable_breakpoint,
            disable=self._disable_breakpoint,
        )

    def _remove_breakpoint(self, breakpoint_id: int) -> None:
        registration = self._breakpoints_by_id.get(breakpoint_id)
        if registration is None or registration.removed:
            return
        registration.removed = True
        registration.enabled = False

    def _enable_breakpoint(self, breakpoint_id: int) -> None:
        registration = self._breakpoints_by_id.get(breakpoint_id)
        if registration is None or registration.removed:
            return
        registration.enabled = True

    def _disable_breakpoint(self, breakpoint_id: int) -> None:
        registration = self._breakpoints_by_id.get(breakpoint_id)
        if registration is None or registration.removed:
            return
        registration.enabled = False

    def _consume_pause_request(self) -> bool:
        pause_requested = self._pause_requested_this_scan
        self._pause_requested_this_scan = False
        return pause_requested

    def _peek_live_tag_value(self, name: str, default: Any) -> Any:
        """Read a tag as seen by live Tag.value access."""
        has_override, value = self._input_overrides.get_live_override(name)
        if has_override:
            return value

        resolved, value = self._system_runtime.resolve(name, self._state)
        if resolved:
            return value

        return self._state.tags.get(name, default)

    def _calculate_dt(self) -> float:
        """Calculate scan delta time based on current time mode."""
        if self._time_mode == TimeMode.REALTIME:
            now = time.perf_counter()
            if self._last_step_time is None:
                self._last_step_time = now
            dt = now - self._last_step_time
            self._last_step_time = now
            return dt
        return self._dt

    def _prepare_scan(self) -> tuple[ScanContext, float]:
        """Create and initialize scan context before logic evaluation."""
        ctx = ScanContext(
            self._state,
            resolver=self._system_runtime.resolve,
            read_only_tags=self._system_runtime.read_only_tags,
        )

        self._system_runtime.on_scan_start(ctx)
        self._input_overrides.apply_pre_scan(ctx)

        dt = self._calculate_dt()
        ctx.set_memory("_dt", dt)
        return ctx, dt

    def _capture_previous_states(self, ctx: ScanContext) -> None:
        """Batch _prev:* updates used by edge detection conditions."""
        for name in self._state.tags:
            ctx.set_memory(f"_prev:{name}", ctx.get_tag(name))
        for name in ctx._tags_pending:
            if name not in self._state.tags:
                ctx.set_memory(f"_prev:{name}", ctx.get_tag(name))

    def _commit_scan(self, ctx: ScanContext, dt: float) -> None:
        """Finalize one scan and commit all batched writes."""
        previous_state = self._state
        previous_tip_scan_id = previous_state.scan_id
        self._input_overrides.apply_post_logic(ctx)

        self._capture_previous_states(ctx)
        self._system_runtime.on_scan_end(ctx)
        self._state = ctx.commit(dt=dt)
        evicted_scan_ids = self._history._append(self._state)
        for evicted_scan_id in evicted_scan_ids:
            self._rung_traces_by_scan.pop(evicted_scan_id, None)
            if (
                self._latest_committed_trace_event is not None
                and self._latest_committed_trace_event[0] == evicted_scan_id
            ):
                self._latest_committed_trace_event = None

        # Keep playhead following newest state unless manually moved.
        if self._playhead == previous_tip_scan_id:
            self._playhead = self._state.scan_id

        if not self._history.contains(self._playhead):
            self._playhead = self._history.oldest_scan_id

        self._evaluate_monitors(previous_state=previous_state, current_state=self._state)
        self._evaluate_breakpoints(state=self._state)
        self._sync_runtime_flags_from_state()

    def _evaluate_monitors(
        self, *, previous_state: SystemState, current_state: SystemState
    ) -> None:
        for monitor_id in sorted(self._monitors_by_id):
            registration = self._monitors_by_id[monitor_id]
            if registration.removed or not registration.enabled:
                continue

            previous_value = previous_state.tags.get(registration.tag_name)
            current_value = current_state.tags.get(registration.tag_name)
            if current_value != previous_value:
                registration.callback(current_value, previous_value)

    def _evaluate_breakpoints(self, *, state: SystemState) -> None:
        self._pause_requested_this_scan = False
        for breakpoint_id in sorted(self._breakpoints_by_id):
            registration = self._breakpoints_by_id[breakpoint_id]
            if registration.removed or not registration.enabled:
                continue

            if not registration.predicate(state):
                continue

            if registration.action == "snapshot":
                assert registration.label is not None
                self._history._label_scan(
                    registration.label,
                    state.scan_id,
                    metadata=self._snapshot_metadata_for_state(state),
                )
                continue

            self._pause_requested_this_scan = True

    def _snapshot_metadata_for_state(self, state: SystemState) -> dict[str, Any]:
        rtc_now = self._rtc_at_sim_time(state.timestamp)
        wall_now = self._normalize_rtc_datetime(datetime.now())
        return {
            "rtc_iso": rtc_now.isoformat(),
            "rtc_offset_seconds": float((rtc_now - wall_now).total_seconds()),
        }

    def scan_steps(self) -> Generator[tuple[int, Rung, ScanContext], None, None]:
        """Execute one scan cycle and yield after each rung evaluation.

        Scan phases:
        1. Create ScanContext from current state
        2. Apply pending patches to context
        3. Apply persistent force overrides (pre-logic)
        4. Calculate dt and inject into context
        5. Evaluate all logic (writes batched in context), yielding after each rung
        6. Re-apply force overrides (post-logic)
        7. Batch _prev:* updates for edge detection
        8. Commit all changes in single operation

        The commit in phase 8 only happens when the generator is exhausted.
        """
        self._ensure_running()
        ctx, dt = self._prepare_scan()

        # Evaluate logic rung-by-rung, yielding at rung boundaries.
        for i, rung in enumerate(self._logic):
            rung.evaluate(ctx)
            yield i, rung, ctx

        self._commit_scan(ctx, dt)

    def scan_steps_debug(self) -> Generator[ScanStep, None, None]:
        """Execute one scan cycle and yield ``ScanStep`` objects at all boundaries.

        Yields a `ScanStep` at each:

        - Top-level rung boundary (``kind="rung"``)
        - Branch entry / exit (``kind="branch"``)
        - Subroutine call and body steps (``kind="subroutine"``)
        - Individual instruction boundaries (``kind="instruction"``)

        Each `ScanStep` carries source location metadata (``source_file``,
        ``source_line``, ``end_line``), rung enable state, and a trace of
        evaluated conditions and instructions.

        This is the API used by the DAP adapter.  Prefer `scan_steps()` for
        non-debug consumers — it has less overhead and a simpler yield type.

        Note:
            Like `scan_steps()`, the scan is committed only when the generator
            is **fully exhausted**.
        """
        self._ensure_running()
        pending_scan_id = self._state.scan_id + 1
        events_by_rung: dict[int, list[RungTraceEvent]] = {}
        self._start_inflight_debug_scan(scan_id=pending_scan_id, events_by_rung=events_by_rung)

        try:
            for step in self._debugger.scan_steps_debug(self):
                event = self._rung_trace_event_from_step(step)
                events_by_rung.setdefault(step.rung_index, []).append(event)
                self._latest_inflight_trace_event = (pending_scan_id, step.rung_index, event)
                yield step

            scan_id = self._state.scan_id
            self._rung_traces_by_scan[scan_id] = {
                rung_id: RungTrace(scan_id=scan_id, rung_id=rung_id, events=tuple(events))
                for rung_id, events in sorted(events_by_rung.items())
            }
            if self._latest_inflight_trace_event is not None:
                _inflight_scan_id, latest_rung_id, latest_event = self._latest_inflight_trace_event
                self._latest_committed_trace_event = (scan_id, latest_rung_id, latest_event)
        finally:
            self._clear_inflight_debug_scan()

    def _start_inflight_debug_scan(
        self,
        *,
        scan_id: int,
        events_by_rung: dict[int, list[RungTraceEvent]],
    ) -> None:
        self._inflight_scan_id = scan_id
        self._inflight_rung_events = events_by_rung
        self._latest_inflight_trace_event = None

    def _clear_inflight_debug_scan(self) -> None:
        self._inflight_scan_id = None
        self._inflight_rung_events = {}
        self._latest_inflight_trace_event = None

    def _rung_trace_event_from_step(self, step: ScanStep) -> RungTraceEvent:
        return RungTraceEvent(
            kind=step.kind,
            source_file=step.source_file,
            source_line=step.source_line,
            end_line=step.end_line,
            subroutine_name=step.subroutine_name,
            depth=step.depth,
            call_stack=step.call_stack,
            enabled_state=step.enabled_state,
            instruction_kind=step.instruction_kind,
            trace=step.trace,
        )

    def prepare_scan(self) -> tuple[ScanContext, float]:
        """Debugger-facing scan preparation API."""
        return self._prepare_scan()

    def commit_scan(self, ctx: ScanContext, dt: float) -> None:
        """Debugger-facing scan commit API."""
        self._commit_scan(ctx, dt)

    def iter_top_level_rungs(self) -> Iterable[Rung]:
        """Debugger-facing top-level rung iterator."""
        return self._logic

    def evaluate_condition_value(
        self,
        condition: Any,
        ctx: ScanContext | ConditionView,
    ) -> tuple[bool, list[dict[str, Any]]]:
        """Debugger-facing condition evaluation API."""
        return self._evaluate_condition_value(condition, ctx)

    def condition_term_text(self, condition: Any, details: list[dict[str, Any]]) -> str:
        """Debugger-facing condition summary API."""
        return self._condition_term_text(condition, details)

    def condition_annotation(self, *, status: str, expression: str, summary: str) -> str:
        """Debugger-facing annotation API."""
        return self._condition_annotation(
            status=status,
            expression=expression,
            summary=summary,
        )

    def condition_expression(self, condition: Any) -> str:
        """Debugger-facing expression rendering API."""
        return self._condition_expression(condition)

    def _evaluate_condition_value(
        self,
        condition: Any,
        ctx: ScanContext | ConditionView,
    ) -> tuple[bool, list[dict[str, Any]]]:
        return self._condition_trace.evaluate(condition, ctx)

    def _condition_term_text(self, condition: Any, details: list[dict[str, Any]]) -> str:
        return self._condition_trace.summary(condition, details)

    def _condition_annotation(self, *, status: str, expression: str, summary: str) -> str:
        return self._condition_trace.annotation(
            status=status,
            expression=expression,
            summary=summary,
        )

    def _condition_expression(self, condition: Any) -> str:
        return self._condition_trace.expression(condition)

    def step(self) -> SystemState:
        """Execute one full scan cycle and return the committed state."""
        self._ensure_running()
        return self._run_single_scan(consume_pause_request=True)

    def _run_single_scan(self, *, consume_pause_request: bool) -> SystemState:
        for _ in self.scan_steps():
            pass

        if consume_pause_request:
            self._consume_pause_request()
        return self._state

    def run(self, cycles: int) -> SystemState:
        """Execute up to ``cycles`` scans, stopping early on pause breakpoints.

        Args:
            cycles: Number of scans to execute.

        Returns:
            The final SystemState after all cycles.
        """
        self._ensure_running()
        for _ in range(cycles):
            self._consume_pause_request()
            self._run_single_scan(consume_pause_request=False)
            if self._consume_pause_request():
                break
        return self._state

    def run_for(self, seconds: float) -> SystemState:
        """Run until simulation time advances by N seconds or a pause breakpoint fires.

        Args:
            seconds: Minimum simulation time to advance.

        Returns:
            The final SystemState after reaching the target time.
        """
        self._ensure_running()
        target_time = self._state.timestamp + seconds
        while self._state.timestamp < target_time:
            self._consume_pause_request()
            self._run_single_scan(consume_pause_request=False)
            if self._consume_pause_request():
                break
        return self._state

    def run_until(
        self,
        *conditions: Condition | Tag | tuple[Condition | Tag, ...] | list[Condition | Tag],
        max_cycles: int = 10000,
    ) -> SystemState:
        """Run until condition is true, pause breakpoint fires, or max_cycles reached.

        Args:
            conditions: ``Tag`` / ``Condition`` expressions evaluated with implicit AND.
            max_cycles: Maximum scans before giving up (default 10000).

        Returns:
            The state that matched the condition, or final state if max reached.
        """
        predicate = self._compile_condition_predicate(*conditions, method="run_until")
        return self.run_until_fn(predicate, max_cycles=max_cycles)

    def run_until_fn(
        self,
        predicate: Callable[[SystemState], bool],
        *,
        max_cycles: int = 10000,
    ) -> SystemState:
        """Run until callable predicate is true, paused, or ``max_cycles`` reached.

        Args:
            predicate: Callable receiving committed ``SystemState`` snapshots.
            max_cycles: Maximum scans before giving up (default 10000).

        Returns:
            The state that matched the predicate, or final state if max reached.
        """
        self._ensure_running()
        for _ in range(max_cycles):
            self._consume_pause_request()
            self._run_single_scan(consume_pause_request=False)
            pause_requested = self._consume_pause_request()
            if predicate(self._state) or pause_requested:
                break
        return self._state



"""Public pyrung API facade.

Import user-facing DSL symbols from this module:

    from pyrung import Bool, Program, Rung, PLCRunner
"""

from pyrung.core import (
    Block,
    Bool,
    Char,
    Dint,
    Field,
    InputBlock,
    Int,
    ModbusAddress,
    ModbusRtuTarget,
    ModbusTcpTarget,
    OutputBlock,
    PLCRunner,
    Program,
    RangeComparison,
    RangeSlotView,
    Real,
    RegisterType,
    Rung,
    SlotView,
    TagType,
    Td,
    Th,
    TimeMode,
    TimeUnit,
    Tm,
    Tms,
    Ts,
    Word,
    WordOrder,
    all_of,
    any_of,
    auto,
    blockcopy,
    branch,
    calc,
    call,
    comment,
    copy,
    count_down,
    count_up,
    event_drum,
    fall,
    fill,
    forloop,
    immediate,
    latch,
    named_array,
    off_delay,
    on_delay,
    out,
    pack_bits,
    pack_text,
    pack_words,
    program,
    receive,
    reset,
    return_early,
    rise,
    run_enabled_function,
    run_function,
    search,
    send,
    shift,
    subroutine,
    system,
    time_drum,
    to_ascii,
    to_binary,
    to_text,
    to_value,
    udt,
    unpack_to_bits,
    unpack_to_words,
)

__all__ = [
    "PLCRunner",
    "system",
    "TimeMode",
    "TimeUnit",
    "Tms",
    "Ts",
    "Tm",
    "Th",
    "Td",
    "Field",
    "auto",
    "udt",
    "named_array",
    "TagType",
    "Bool",
    "Int",
    "Dint",
    "Real",
    "Char",
    "Word",
    "Block",
    "InputBlock",
    "OutputBlock",
    "RangeComparison",
    "RangeSlotView",
    "SlotView",
    "Program",
    "Rung",
    "program",
    "branch",
    "comment",
    "forloop",
    "immediate",
    "subroutine",
    "out",
    "latch",
    "reset",
    "copy",
    "run_function",
    "run_enabled_function",
    "blockcopy",
    "fill",
    "pack_bits",
    "pack_text",
    "pack_words",
    "unpack_to_bits",
    "unpack_to_words",
    "calc",
    "call",
    "return_early",
    "count_up",
    "count_down",
    "event_drum",
    "search",
    "shift",
    "on_delay",
    "off_delay",
    "time_drum",
    "send",
    "receive",
    "ModbusAddress",
    "ModbusRtuTarget",
    "ModbusTcpTarget",
    "RegisterType",
    "WordOrder",
    "rise",
    "fall",
    "all_of",
    "any_of",
    "to_value",
    "to_ascii",
    "to_text",
    "to_binary",
]
